#!/usr/bin/env python
# coding: utf-8

# In[1]:


import numpy as np
import astropy
import astropy.constants as constants
from astropy.table import Table
from astropy import units as u
import pickle

#Global Imports
import matplotlib 
import matplotlib.pyplot as plt
matplotlib.rcParams['font.family']= 'Times New Roman'
matplotlib.rcParams['font.sans-serif']= ['Bitstream Vera Sans']
matplotlib.rcParams['text.usetex']= True
matplotlib.rcParams['mathtext.fontset']= 'stixsans'

import seaborn as sns
sns.set_context('talk')
sns.set_style('ticks')
sns.set_palette('colorblind')
sns.set_palette('colorblind')
cs=sns.color_palette('colorblind')

c=constants.c.to('km/s').value

def sph2vec(ra,dec):
    return np.array([np.sin(np.pi/2.-dec)*np.cos(ra),np.sin(np.pi/2.-dec)*np.sin(ra),np.cos(np.pi/2.-dec)])

def zhelio_to_zcmb(ra,dec,z_helio):
    ra_cmb = 167.99*np.pi/180
    dec_cmb = -7.22*np.pi/180
    v_cmb = 369.
    z_gal_cmb = v_cmb*np.dot(sph2vec(ra_cmb,dec_cmb),sph2vec(ra,dec))/c
    return (1.+z_helio)*(1.+z_gal_cmb)-1.


# In[ ]:


GLADE = Table.read("GLADE_2.4.txt",format='ascii')

print("GLADE has "+str(len(GLADE))+" objects.")


# In[ ]:


# remove galaxies with no reported redshift
no_zs = GLADE['col11'] == 'null'
GLADE.remove_rows(no_zs)

print("GLADE has "+str(len(GLADE))+" galaxies with a measured redshift.")


# In[4]:


# remove QSOs and galaxy clusters.
no_clusters = GLADE['col6'] == 'C'
GLADE.remove_rows(no_clusters)

no_QSOs = GLADE['col6'] == 'Q'
GLADE.remove_rows(no_QSOs)

print("GLADE has "+str(len(GLADE))+" galaxies.")


# In[5]:


# Apply CMB frame correction
ra = GLADE['col7']*np.pi/180.
dec = GLADE['col8']*np.pi/180.
z = GLADE['col11']

ra = ra.astype('float64')
dec = dec.astype('float64')
z = z.astype('float64')

zcmb = zhelio_to_zcmb(ra,dec,z)

GLADE.add_column(zcmb,name='zcmb')

# Make sure redshifts are positive
mask = GLADE['zcmb'] < 0
GLADE.remove_rows(mask)
print("GLADE has "+str(len(GLADE))+" galaxies.")


# In[6]:


# Assign redshift errors to spec, photo, and hyperleda galaxies
zerror_sp = 1.5e-4
zerror_ph = 1.5e-2
zerror_hyperleda = 0.36

zs_sp = GLADE['col21'] == 3
zs_ph = GLADE['col21'] == 1
zs_hyperleda = (GLADE['col21'] == 2) & (GLADE['col3'] != 'null')

zerrors = np.zeros(len(GLADE))
GLADE.add_column(zerrors,name='zerror')

z_sp = GLADE['zcmb'][zs_sp]
z_ph = GLADE['zcmb'][zs_ph]
z_hyperleda = GLADE['zcmb'][zs_hyperleda]

z_sp = z_sp.astype('float64')
z_ph = z_ph.astype('float64')
z_hyperleda = z_hyperleda.astype('float64')

GLADE['zerror'][zs_sp] = zerror_sp*np.ones_like(z_sp)
GLADE['zerror'][zs_ph] = zerror_ph*np.ones_like(z_ph)
GLADE['zerror'][zs_hyperleda] = zerror_hyperleda*z_hyperleda


# In[7]:


# remove GWCC galaxies (distances to redshift gals)
no_errors = GLADE['zerror'] == 0
GLADE.remove_rows(no_errors)

print("GLADE has "+str(len(GLADE))+" galaxies.")


# In[8]:


# Select columns of interest
ra = GLADE['col7']*np.pi/180.
dec = GLADE['col8']*np.pi/180.
z = GLADE['zcmb']
B = GLADE['col12']
K = GLADE['col19']
sigmaz = GLADE['zerror']

ra = ra.astype('float64')
dec = dec.astype('float64')
z = z.astype('float64')
B = B.data
K = K.data

mask = B == 'null'
B[mask] = '0'
mask = K == 'null'
K[mask] = '0'

B = B.astype('float64')
K = K.astype('float64')

B[B==0] = np.nan
K[K==0] = np.nan


# In[9]:


import h5py
import healpy as hp

# Save GLADE catalog
bands=['B', 'K']
nGal = len(z)
m = np.column_stack((B,K))

nside = 1024
ind = hp.pixelfunc.ang2pix(nside,dec+np.pi/2,ra)

ra_dec_lim = 0
ra_min = 0.0
ra_max = np.pi*2.0
dec_min = -np.pi/2.0
dec_max = np.pi/2.0 

with h5py.File("glade.hdf5", "w") as f:
    f.create_dataset("ra", data=ra)
    f.create_dataset("dec", data=dec)
    f.create_dataset("z", data=z)
    f.create_dataset("sigmaz", data=sigmaz)
    f.create_dataset("skymap_indices", data=ind)
    f.create_dataset("radec_lim", data=np.array([ra_dec_lim,ra_min,ra_max,dec_min,dec_max]))
    for j in range(len(bands)):
        f.create_dataset("m_{0}".format(bands[j]), data=m[:,j])


# In[18]:


len(np.where(~np.isnan(B)==True)[0])


# In[19]:


len(np.where(~np.isnan(K)==True)[0])


# In[ ]:







"""
LALinference posterior samples class and methods
Ignacio Magana, Ankan Sur
"""
import numpy as np
from scipy.stats import gaussian_kde
from astropy import units as u
import h5py
from .skymap import ra_dec_from_ipix
from ..prior.priors import distance_distribution
import json
import healpy as hp
import copy
import sys

from scipy.interpolate import RegularGridInterpolator

class load_posterior_samples(object):
    """
    Posterior samples class and methods.

    Parameters
    ----------
    posterior_samples : Path to posterior samples file to be loaded.
    field : Internal field of the json or the h5 file
    """
    
    def __init__(self,posterior_samples,field=None):
        self.posterior_samples = posterior_samples

        self.field=field
        self.load_posterior_samples()

    def load_posterior_samples(self):
        """
        Method to handle different types of posterior samples file formats.
        Currently it supports .dat (LALinference), .hdf5 (GWTC-1),
        .h5 (PESummary) and .hdf (pycbcinference) formats.
        """
        if self.posterior_samples[-3:] == 'dat':
            samples = np.genfromtxt(self.posterior_samples, names = True)
           
            self.distance = np.array([var for var in samples['luminosity_distance']])
            self.ra =  np.array([var for var in samples['ra']])
            self.dec =  np.array([var for var in samples['dec']])
            self.mass_1 =  np.array([var for var in samples['mass_1']])
            self.mass_2 =  np.array([var for var in samples['mass_2']])
            self.nsamples = len(self.distance)

        if self.posterior_samples[-4:] == 'hdf5':
            if self.posterior_samples[-11:] == 'GWTC-1.hdf5':
                if self.posterior_samples[-20:] == 'GW170817_GWTC-1.hdf5':
                    dataset_name = 'IMRPhenomPv2NRT_lowSpin_posterior'
                else:
                    dataset_name = 'IMRPhenomPv2_posterior'
                file = h5py.File(self.posterior_samples, 'r')
                data = file[dataset_name]
                self.distance = data['luminosity_distance_Mpc']
                self.ra = data['right_ascension']
                self.dec = data['declination']
                self.mass_1 = data['m1_detector_frame_Msun']
                self.mass_2 = data['m2_detector_frame_Msun']
                self.nsamples = len(self.distance)
                file.close()

        if self.posterior_samples.endswith('.json'):
            with open(self.posterior_samples) as f:
                data = json.load(f)

            PE_struct=data['posterior_samples'][self.field]

            m1_ind=PE_struct['parameter_names'].index('mass_1')
            m2_ind=PE_struct['parameter_names'].index('mass_2')
            dl_ind=PE_struct['parameter_names'].index('luminosity_distance')
            ra_ind=PE_struct['parameter_names'].index('ra')
            dec_ind=PE_struct['parameter_names'].index('dec')
                        
            nsamp=len(PE_struct['samples'])
            
            self.distance = np.array(PE_struct['samples'])[:,dl_ind].reshape(-1)
            self.ra = np.array(PE_struct['samples'])[:,ra_ind].reshape(-1)
            self.dec = np.array(PE_struct['samples'])[:,dec_ind].reshape(-1)
            self.mass_1 = np.array(PE_struct['samples'])[:,m1_ind].reshape(-1)
            self.mass_2 = np.array(PE_struct['samples'])[:,m2_ind].reshape(-1)
            self.nsamples = len(self.distance)

        if self.posterior_samples[-2:] == 'h5':
            file = h5py.File(self.posterior_samples, 'r')

            if self.field is None:
                approximants = ['PublicationSamples','C01:Mixed','C01:PhenomPNRT-HS', 
                                'C01:NRSur7dq4', 'C01:IMRPhenomPv3HM', 'C01:IMRPhenomPv2',
                                'C01:IMRPhenomD', 'C01:IMRPhenomPv2_NRTidal:LowSpin', 
                                'C01:IMRPhenomPv2_NRTidal:HighSpin']
                for approximant in approximants:
                    try:
                        data = file[approximant]
                        print("Using "+approximant+" posterior")
                        break
                    except KeyError:
                        continue
            else:
                data=file[self.field]

            self.distance = data['posterior_samples']['luminosity_distance']
            self.ra = data['posterior_samples']['ra']
            self.dec = data['posterior_samples']['dec']
            self.mass_1 = data['posterior_samples']['mass_1']
            self.mass_2 = data['posterior_samples']['mass_2']
            self.nsamples = len(self.distance)
            file.close()

        if self.posterior_samples[-3:] == 'hdf':
            file = h5py.File(self.posterior_samples, 'r')
            self.distance = file['samples/distance'][:]
            self.ra = file['samples/ra'][:]
            self.dec = file['samples/dec'][:]
            self.mass_1 = file['samples/mass_1'][:]
            self.mass_2 = file['samples/mass_2'][:]
            self.nsamples = len(self.distance)
            file.close()

    def marginalized_sky(self):
        """
        Computes the marginalized sky localization posterior KDE.
        """
        return gaussian_kde(np.vstack((self.ra, self.dec)))


class reweight_posterior_samples(object):
    """
    Posterior samples class and methods.

    Parameters
    ----------
    cosmo : Fast cosmology class
    mass_priors: Fast mass_distributions class
    """
    
    def __init__(self,cosmo,mass_priors):
        self.cosmo = cosmo
        # Prior distribution used in this work
        self.source_frame_mass_prior = mass_priors
    def jacobian_times_prior(self,z):
        """
        (1+z)^2 * ddL/dz * dL^2
        """
        jacobian = np.power(1+z,2)*self.cosmo.ddgw_dz(z)
        dgw = self.cosmo.dgw_z(z)

        return jacobian*(dgw**2)
        
    def compute_source_frame_samples(self, GW_distance, det_mass_1, det_mass_2):
        """
        Posterior samples class and methods.

        Parameters
        ----------
        GW_distance: GW distance samples in Mpc
        det_mass_1, det_mass_2 : detector frame mass samples in Msolar
        H0 : Hubble constant value in kms-1Mpc-1
        """
        redshift = self.cosmo.z_dgw(GW_distance)

        mass_1_source = det_mass_1/(1+redshift)
        mass_2_source = det_mass_2/(1+redshift)
        return redshift, mass_1_source, mass_2_source

    def get_kde(self, data, weights):
        # deal first with the weights
        weights, norm, neff = self.check_weights(weights)
        if norm != 0:
            try:
                kde = gaussian_kde(data, weights=weights)
            except:
                print("KDE problem! create a default KDE with norm=0")
                print("norm: {} -> 0, neff: {}".format(norm,neff))
                norm = 0
                kde = gaussian_kde(data)
        else:
            kde = gaussian_kde(data)
            
        return kde, norm
                    
    def ignore_weights(self, weights):

        weights = np.ones(len(weights))
        norm = 0
        return weights, norm

    def check_weights(self, weights):
        """
        Check the weights values to prevent gaussian_kde crash when Neff <= 1,
        where Neff is an internal variable of gaussian_kde
        defined by Neff = sum(weights)^2/sum(weights^2)
        careful, cases with Neff = 1+2e-16 = 1.0000000000000002
        have been seen and give crash: set Neff limit to >= 2
        """
        neff = 0
        if np.isclose(max(weights),0,atol=1e-50):
            weights, norm = self.ignore_weights(weights)
        else:
            neff = sum(weights)**2/sum(weights**2)
            if neff<2:
                weights, norm = self.ignore_weights(weights)
            else:
                norm = np.sum(weights)/len(weights)
        return weights, norm, neff
    
    def marginalized_redshift_reweight(self, redshift, mass_1_source, mass_2_source):
        """
        Computes the marginalized distance posterior KDE.
        """
        # Re-weight
        weights = self.source_frame_mass_prior.joint_prob(mass_1_source,mass_2_source)/self.jacobian_times_prior(redshift)

        return self.get_kde(redshift,weights)

    def marginalized_redshift(self, redshift):
        """
        Computes the marginalized distance posterior KDE.
        """
        # remove dgw^2 prior and include dz/ddgw jacobian
        weights = 1/(self.cosmo.ddgw_dz(redshift)*self.cosmo.dgw(redshift)**2)
        return self.get_kde(redshift,weights)


class make_pixel_px_function(object):
    """
    Identify the posterior samples which lie within some angular radius
    (depends on skymap pixel size) of the centre of each pixel
    """
    
    def __init__(self, samples, skymap, npixels=30, thresh=0.999):
        """
        Parameters
        ----------
        samples : posterior_samples object
            The GW samples
        skymap : object
            The GW skymap
        npixels : int, optional
            The minimum number of pixels desired to cover given sky area of
            the GW event (default=30)
        thresh : float, optional
            The sky area threshold (default=0.999)
        """
        
        self.skymap = skymap
        self.samples = samples
        nside=1
        indices,prob = skymap.above_percentile(thresh, nside=nside)
    
        while len(indices) < npixels:
            nside = nside*2
            indices,prob = skymap.above_percentile(thresh, nside=nside)
        
        self.nside = nside
        print('{} pixels to cover the {}% sky area (nside={})'.format(len(indices),thresh*100,nside))
        
        dicts = {}
        for i,idx in enumerate(indices):
            dicts[idx] = prob[i]
        self.indices = indices
        self.prob = dicts # dictionary - given a pixel index, returns skymap prob

        
    def identify_samples(self, idx, minsamps=100):
        """
        Find the samples required 
        
        Parameters
        ----------
        idx : int
            The pixel index
        minsamps : int, optional
            The threshold number of samples to reach per pixel
            
        Return
        ------
        sel : array of ints
            The indices of posterior samples for pixel idx
        """
        
        ipix_samples = hp.pixelfunc.ang2pix(self.nside, np.pi/2-self.samples.dec, self.samples.ra, nest=self.skymap.nested)
        sel = np.where(ipix_samples == idx)[0]
        if len(sel) >= minsamps:
            print("{} samples fall in pix {}".format(len(sel),idx))
            return sel

        # not enough samples in pixel 'idx', we need to extend the search
        racent,deccent = ra_dec_from_ipix(self.nside, idx, nest=self.skymap.nested)
        separations = angular_sep(racent,deccent,self.samples.ra,self.samples.dec)
        sep = hp.pixelfunc.max_pixrad(self.nside)/2. # choose initial separation
        step = sep/2. # choose step size for increasing radius
        
        sel = np.where(separations<sep)[0] # find all the samples within the angular radius sep from the pixel centre
        nsamps = len(sel)
        while nsamps < minsamps:
            sep += step
            sel = np.where(separations<sep)[0]
            nsamps = len(sel)
            if sep > np.pi:
                raise ValueError("Problem with the number of posterior samples.")
        print('pixel idx {}: angular radius: {} radians, No. samples: {}'.format(idx,sep,len(sel)))
         
        return sel
        

def identify_samples_from_posterior(ra_los, dec_los, ra, dec, nsamps=1000):
    """
    Find the angular separation between all posterior samples and a specific
    LOS. Return the indices of the nsamps closest samples, as well as the 
    maxiumum separation of those samples.
    
    Parameters
    ----------
    ra_los : float
        right ascension of the line-of-sight (radians)
    dec_los : float
        declination of the line-of-sight (radians)
    ra : array of floats
        right ascensions of a set of samples (radians)
    dec : array of floats
        declinations of a set of samples (radians)
    nsamps : int, optional
        The number of samples to select (default=1000)
        
    Return
    ------
    index : array of ints
        The indices of of the nsamps samples
    ang_rad_max : float
        The maximum angular radius between the selected samples and the LOS
    """

    separations = angular_sep(ra_los,dec_los,ra,dec)
    sep_argsort = np.argsort(separations)
    index = sep_argsort [:nsamps]
    ang_rad_max = separations [nsamps-1]

    return index, ang_rad_max

        
    
def angular_sep(ra1,dec1,ra2,dec2):
    """Find the angular separation between two points, (ra1,dec1)
    and (ra2,dec2), in radians."""
    
    cos_angle = np.sin(dec1)*np.sin(dec2) + np.cos(dec1)*np.cos(dec2)*np.cos(ra1-ra2)
    angle = np.arccos(cos_angle)
    return angle



# Global Imports
import numpy as np
from scipy.integrate import cumtrapz
from scipy.optimize import fmin
from scipy.interpolate import interp1d, UnivariateSpline


class confidence_interval(object):
    def __init__(self, posterior, param, level=0.683, verbose=False):
        self.posterior = posterior
        self.param = param
        self.level = level
        self.verbose = verbose
        self.lower_level, self.upper_level = self.HDI()
        self.interval = self.upper_level - self.lower_level
        self.map = self.MAP()
        
    def HDI(self):
        cdfvals = cumtrapz(self.posterior, self.param)
        sel = cdfvals > 0.
        x = self.param[1:][sel]
        cdfvals = cdfvals[sel]
        ppf = interp1d(cdfvals, x, fill_value=0., bounds_error=False)

        def intervalWidth(lowTailPr):
            ret = ppf(self.level + lowTailPr) - ppf(lowTailPr)
            if (ret > 0.):
                return ret
            else:
                return 1e4
        HDI_lowTailPr = fmin(intervalWidth, 1. - self.level, disp=self.verbose)[0]
        return ppf(HDI_lowTailPr), ppf(HDI_lowTailPr + self.level)


    def MAP(self):
        sp = UnivariateSpline(self.param, self.posterior, s=0.)
        x_highres = np.linspace(self.param[0], self.param[-1], 100000)
        y_highres = sp(x_highres)
        return x_highres[np.argmax(y_highres)]
        
def str2bool(v):
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')



"""
Injections module

Benoit Revenu
Christos Karathanasis
"""

import bilby as bl
import numpy as np
import pickle 
from tqdm import tqdm
import lalsimulation as lalsim
import os
import logging
from gwcosmo.injections import injections_at_detector, injections_at_source
from bilby.gw import utils as gwutils
import multiprocess as mp
import sys
import time
import copy
import tempfile
from gwcosmo.utilities.cosmology import standard_cosmology
import random
import string
import scipy.integrate as integrate
import h5py

class Counter(object):
    '''
    This class defines a counter in shared memory for usage in the children processes.
    In the current version of the code, where now each child process has its dedicated number of injections
    to compute, the counter object is used only for information; it's not part anymore of the process
    of controlling the number of injections to compute.
    The lock mechanism avoid problems when 2 processes want access at the same time to the same variable.

    Parameters
    ----------
    initval: int
    this is the initial value of the counter (default = -1)

    maxval: int
    this is the maximum value of the counter (default = -1)
    '''

    def __init__(self, initval = -1, maxval = -1):
        self.val = mp.Value('i',initval)
        self.maxval = mp.Value('i',maxval)
        self.lock = mp.Lock()

    def set_value(self, value):
        with self.lock:
            self.val = mp.Value('i',value)
            
    def increment(self,to_add=1):
        with self.lock:
            self.val.value += to_add
            return self.val.value

    def get_value(self):
        with self.lock:
            return self.val.value

    def max_value(self):
        with self.lock:
            return self.maxval.value

def get_dLmax_params(snr):
    '''
    This function returns the values of the parameters of the function dLmax(m1) for snrs 9, 10, 11, 12.
    Parameters are: a[0], a[1], a[2], a[3] with dLmax(m1) = (a[0]+a[1]*m1)*exp(-(m1+a[2])**2/(2*a[3]**2))
    '''
    dLmax_m1 = {}
    if snr == 9:
        dLmax_m1['O1'] = [ 84.31804818, 192.53468201, 182.93373141, 299.50877694]
        dLmax_m1['O2'] = [147.91817378, 175.75506129,  35.93831714, 263.5630744 ]
        dLmax_m1['O3'] = [213.84817574, 244.51138966,  69.23433534, 271.4468494 ]
        dLmax_m1['O4high'] = [436.18835656, 366.88443749,  93.34706107, 343.16794722]
        dLmax_m1['O4low'] = [489.22000474, 834.02779856, 633.62937608, 496.61871247]
    elif snr == 10:
        dLmax_m1['O1'] = [ 86.99423443, 159.58449504, 178.84598956, 300.5293686 ]
        dLmax_m1['O2'] = [128.05686148, 154.41689787,  64.73827621, 280.0737362 ]
        dLmax_m1['O3'] = [160.81115278, 234.71142174, 130.4708518,  294.36698715]
        dLmax_m1['O4high'] = [359.47412722, 313.11598171,  70.02679103, 325.7867842 ]
        dLmax_m1['O4low'] = [353.02830031, 521.1612163,  471.41373238, 457.18316462]
    elif snr == 11:
        dLmax_m1['O1'] = [ 80.00135878, 129.34914561, 140.42625102, 286.23857801]
        dLmax_m1['O2'] = [114.27477259, 130.78112051,  48.1232669,  277.0573754 ]
        dLmax_m1['O3'] = [137.31083208, 204.85159489, 125.53551322, 291.99014915]
        dLmax_m1['O4high'] = [310.99831975, 273.55408815,  73.49104502, 328.08351715]
        dLmax_m1['O4low'] = [245.81416194, 391.50469287, 373.16419088, 424.40083527]
    elif snr == 12:
        dLmax_m1['O1'] = [ 67.88545626, 109.03198868, 111.82572526, 274.54170286]
        dLmax_m1['O2'] = [102.23016472, 113.16162086,  24.72815689, 267.31760986]
        dLmax_m1['O3'] = [119.23589567, 182.81133974, 129.51557969, 294.25833901]
        dLmax_m1['O4high'] = [274.3325264,  268.83221473, 165.03056849, 371.29597527]
        dLmax_m1['O4low'] = [180.99809457, 273.19457205, 200.29022794, 348.78256775]
    else:
        print("dLmax(m1) parameters not available for snr {}.".format(snr))
    return dLmax_m1

        
class Create_injections(object):
    '''
    This is the class dealing with injections in the detector frame only. The source frame case will be considered later.

    Parameters
    ----------
    
    asd_path: str
    path of the strain files (2 columns: frequency - amplitude), this is NOT the psd

    psd_opts: str
    is either 'low' or 'high' corresponding to the pessimistic or optimistic (respectively) sentivity for O4

    days_of_run: dict
    number of days of run for O1, O2, O3
    default = {'O3':330,'O2':268,'O1':129}

    duty_factors: dict
    duty factors for O1, O2, O3
    default = {'O4':{'H1':0.75,'L1':0.75,'V1':0.75},
               'O3':{'H1':0.7459,'L1':0.7698,'V1':0.7597},
               'O2':{'H1':0.653,'L1':0.618,'V1':0.0777},
               'O1':{'H1':0.646,'L1':0.574,'V1':-1}}
    in https://arxiv.org/pdf/1304.0670.pdf
    number of days of O1: 12 September 2015 - 19 January 2016 with duty factors 64.6% (H) and 57.4% (L), i.e. 129 days
    number of days of O2: 30 November 2016 - 25 August 2017 with duty factors 65.3% (H) and 61.8% (L), i.e. 268 days
    Virgo/O2 was in data taking mode during 24.5 days (2017-08-01:10:00:00 - 2017-08-25:22:00:00) with a duty cycle of 85.08%.
    So that during O2, Virgo must be considered online with a probability of 85.08% * 24.5 / 268 = 7.77%
    Reference for Virgo/O2: https://wiki.virgo-gw.eu/DataAnalysis/DetChar/DetCharO2Summary and https://logbook.virgo-gw.eu/virgo/?r=39301
    for O3, number of days = 330 days (28519200/86400) values are here: https://wiki.ligo.org/Operations/O3OfficialNumbers

    frame: str
    reference frame for masses, either 'detectors_frame' or 'source_frame'
    default = 'detectors_frame'

    nprocs: int
    this is the number of child processes started by the main process to compute injections

    tmp_to_dict: str
    name of a temporary file
    the temporary file will be converted into a dict
    no injections are computed if activated

    tmp_to_stdout: str
    name of a temporary file
    the contents of the temporary file is dumped on stdout

    tmpfile_to_dict: str
    name of a file containing the list of temporary files (full path) containing injections
    this is used when using a large number of temporary files coming from separate computing centers

    approx: str
    model used for the theoretical signal in injections
    default: 'IMRPhenomPv2'

    priors: str
    this is the path to a file containing a prior for Bilby
    this prior file is sent to the function bilby.core.prior.dict.PriorDict
    default: None i.e. we create the prior in the code

    fmin: float
    this is the value of the frequency in Hz used in gwutils.calculate_time_to_merger(frequency = fmin,...)
    and also in waveform_arguments = dict(.. reference_frequency = fmin)
    default: 20 (Hz)

    sampling frequency: float
    sampling frequency of the signal in Hz
    this value is used in bilby.gw.waveform_generator.WaveformGenerator(... sampling_frequency = self.fsamp...)
    default: 4096 Hz

    dLmax_depends_on_m1: bool
    activate or not the dependence of dLmax on m1
    if not activated, dLmax is constant with m1, with a constant value depending of O1, O2, O3
    default: True (i.e. use a dLmax(m1) with SNR_thres=9)

    SNR_thres: float
    the minimum value of the SNR for detected injections
    default: 9
    '''
    
    def __init__(self,detectors=['H1', 'L1', 'V1'],
                 psd=None,
                 psd_opts=None,
                 asd_path='./',
                 days_of_runs=None,
                 duty_factors=None,
                 frame='detectors_frame',
                 nprocs=1,
                 tmp_to_dict=None,
                 tmp_to_stdout=None,
                 tmpfile_to_dict=None,
                 approx='IMRPhenomPv2',
                 priors=None,
                 priors_limits=None,
                 fmin=20,
                 sampling_frequency=4096,
                 dLmax_depends_on_m1=False,
                 dLmax_m1_params=None,
                 SNR_thres=9,
                 dump_inj_period=200):


        ###########################################################################
        # define strings and data structure for the communication with the
        # communication process (write_injection function, using a queue)
        self.kill_msg = "CaN YoU terMinate YOURSELF PLEASE?"
        self.header_msg = "header"
        self.footer_msg = "footer"
        self.injection_msg = "NewInjection"
        # we define the list sent by the child injection processes to the write_injections process
        # a single item in the list: the injections
        self.empty = [0]
        ###########################################################################


        logging.disable(logging.INFO)
        if priors != None:
            print("The current version of the injection code does not treat priors from file. Exiting.")
            sys.exit()
                
        if tmp_to_dict != None:
            # convert a temporary injections file into a dict file
            print("Converting temp file {}".format(tmp_to_dict))
            if os.path.isfile(tmp_to_dict):
                self.dict_detected, ninj, toskip = convert_temp_file_to_dict(tmp_to_dict,
                                                                             self.header_msg,
                                                                             self.footer_msg,
                                                                             self.injection_msg)
                if not toskip:
                    dn = os.path.dirname(tmp_to_dict)
                    if dn == '': dn = "./"
                    self.output_file_dict = dn+"/dict_"+os.path.basename(tmp_to_dict)
                    write_object_to_file(self.dict_detected,self.output_file_dict)
                    print("Dict written to file {}".format(self.output_file_dict))
                else:
                    print("Tmp file {} skipped".format(tmp_to_dict))
            else:
                print("Could not open file {}. Exiting.".format(list_to_dict))
            sys.exit()

        if tmpfile_to_dict != None:
            # merge many temporary injections file into a single dict, used in the case of injections from different clusters
            print("Merging tmp files into a single dict.")
            if os.path.isfile(tmpfile_to_dict):
                self.dict_detected = merge_tmpfile(tmpfile_to_dict,
                                                   self.header_msg,
                                                   self.footer_msg,
                                                   self.injection_msg)
            else:
                print("Could not open file {}. Exiting.".format(tmpfile_to_dict))
            sys.exit()
            
        if tmp_to_stdout != None:
            # display on stdout the contents of a temporary injections file
            print("Displaying tmp file {} on stdout".format(tmp_to_stdout))            
            if os.path.isfile(tmp_to_stdout):
                try:
                    fd = open(tmp_to_stdout,"rb")
                except Exception as e:
                    print(e)
                    print("Could not open file {}. Exiting.".format(tmp_to_stdout))
                    sys.exit()
                try:
                    evt_list = pickle.load(fd)
                    for i in evt_list:
                        print(i)
                except Exception as e:
                    print(e)
                    print("Could not unpickle the file {}, file could be damaged, try to redownload it?\nSkipping file...".format(tmp_to_stdout))                
            else:
                print("Invalid file {}. Exiting.".format(tmp_to_stdout))
            sys.exit()
        
        if days_of_runs == None:
            # use the same values as the default ones in gwcosmo/utilities/arguments.py
            days_of_runs = {'O4':330,'O3':330,'O2':268,'O1':129 } # assume O4 days = O3 days
            
        self.total_days = 0
        self.frame = frame
        self.SNR_th = SNR_thres # SNR threshold

        if self.frame != "detectors_frame":
            print("Injections are computed only in the detector frame. Source frame not implemented yet. Exiting.")
            sys.exit()
            
        self.nprocs = nprocs
        print("Machine with {} cores. Will use {} processes.".format(mp.cpu_count(),self.nprocs))
        self.nclients = self.nprocs+1 # +1 for the main process
        self.manager = mp.Manager()
        self.shared_queue = mp.Queue() # prepare the communication pipe
        self.H0_ref = -1 # init, updated later on if frame = source
        self.Om0_ref = -1 # init, updated later on if frame = source

        # self.frame == 'detectors_frame':
        self.generate_priors_G = self.generate_priors_detector_G
        self.generate_priors_R = self.generate_priors_detector_R

        if psd == None:
            self.psd_dict = {'O1':{},'O2':{},'O3':{},'O4':{}}
        else:
            if isinstance(psd,list):
                self.psd_dict ={psd[0]:{}}
            else:
                self.psd_dict ={psd:{}}

        for key in self.psd_dict:
            self.total_days += days_of_runs[key]
        if self.total_days == 0:
            print("Observation time Tobs = 0. Check parameter 'days_of_runs'. Exiting.")
            sys.exit()

        if duty_factors == None:
            # use the same values as the default ones in gwcosmo/utilities/arguments.py
            duty_factors = {'O4':{'H1':0.75,'L1':0.75,'V1':0.75},
                            'O3':{'H1':0.7459,'L1':0.7698,'V1':0.7597},
                            'O2':{'H1':0.653,'L1':0.618,'V1':0.0777},
                            'O1':{'H1':0.646,'L1':0.574,'V1':-1}}
            print("No duty factors provided, using default values: {}".format(duty_factors))
            
        # get proba of drawing O1, O2, O3, O4
        self.prob_of_run = {}#'O4':0,'O3':0,'O2':0,'O1':0}
        ptot = 0
        for key in self.psd_dict:
            if days_of_runs[key] > 0:
                self.prob_of_run[key] = days_of_runs[key]/self.total_days
                ptot += self.prob_of_run[key]
        if abs(1-ptot) > 0.01:
            print("Anomaly: probabilities don't add-up to 1 (ptot = {}), check the days of activity of runs! Exiting.".format(ptot))
            sys.exit()
        else: # rescale prob of runs, to reach exactly 1
            print(self.prob_of_run)
            for i in self.prob_of_run:
                self.prob_of_run[i] /= ptot
            print(self.prob_of_run)

        self.active_runs = list(self.prob_of_run.keys()) # will be use at some places
        
        self.Tobs_year = self.total_days/365.25



        # create bilby PSDs once for all
        self.bilby_psds = {k:{} for k in self.active_runs}
        
        self.psd_opts = psd_opts # copy optional run keyword (psd_opts is 'low' or 'high' for O4)
        if ('O4' in self.active_runs) and (self.psd_opts == None):
            print("O4 activated but no sensitivity is specified. Selecting the 'low' configuration (pessimistic).")
            self.psd_opts = 'low'
            
        nfiles_ok = 0
        for LVCrun in self.active_runs:
            print(LVCrun)
            for ifo in detectors:
                try:
                    # load strain data, these files have 2 columns: frequency - ASD
                    if LVCrun == 'O4':
                        asd_file = asd_path+ifo+'_'+LVCrun+self.psd_opts+'_strain.txt'
                    else:
                        asd_file = asd_path+ifo+'_'+LVCrun+'_strain.txt'
                    print(asd_file)
                    data = np.genfromtxt(asd_file)
                    self.psd_dict[LVCrun][ifo] = {}
                    self.psd_dict[LVCrun][ifo]['frequency'] = data[:,0]
                    self.psd_dict[LVCrun][ifo]['psd'] = data[:,1]**2 # data[:,1] is asd, so **2 for psd
                    self.psd_dict[LVCrun][ifo]['duty_cycle'] = duty_factors[LVCrun][ifo]
                    self.bilby_psds[LVCrun][ifo] = bl.gw.detector.PowerSpectralDensity(frequency_array=self.psd_dict[LVCrun][ifo]['frequency'],
                                                                                       psd_array=self.psd_dict[LVCrun][ifo]['psd'])
                    nfiles_ok += 1
                except Exception as e:
                    print('Problem in loading asd file for  {:},{:}. Setting up without it.'.format(LVCrun,ifo))
                    print(e)

        if nfiles_ok == 0:
            print("No asd file loaded. Check paths! Exiting.")
            sys.exit()
        
        # create object detector_config
        # this object if used when randomly drawing the run (O1, O2, O3, O4) and the online interferometers
        self.detector_config = detector_config(self.prob_of_run,self.psd_dict)

        # copy data from function arguments
        self.dLmax_depends_on_m1 = dLmax_depends_on_m1
        self.dLmax_m1_params_args = dLmax_m1_params # copied from function arguments, obtained from simulations
        self.approx = approx
        self.priors = priors
        self.priors_limits = priors_limits
        self.fmin = fmin
        self.fsamp = sampling_frequency
        self.dump_inj_period = dump_inj_period
        if self.priors_limits == None:
            # use the same values as the default ones in gwcosmo/bin/create_injections
            self.priors_limits = {'Mmin_det':1.,
                                  'Mmax_det':500.,
                                  'power_index_m1_det':-2.,
                                  'power_index_m2_det':1.,
                                  'Mmin_source':2.,
                                  'Mmax_source':300.,
                                  'power_index_m1_source':-2.,
                                  'power_index_m2_source':1.,
                                  'zmin':0.001,
                                  'zmax':4,
                                  'z_alpha':2.,
                                  'power_index_dL':2.,
                                  'dLmin':0.1,
                                  'dLmax':80e3} # same value for all runs

        self.dl_m1_functions =  self.get_condition_func_dl_m1()

        # check if we use dLmax(m1), available for SNRth = 9, 10, 11, 12
        if self.dLmax_depends_on_m1 and ((self.SNR_th != 9) and (self.SNR_th != 10) and (self.SNR_th != 11) and (self.SNR_th != 12)):
            print("Using dLmax(m1) is OK for SNR_th = 9, 10, 11 or 12 only. Exiting.")
            sys.exit()
        
        self.set_priors() # here we set the priors in the global params space and in the reduced one if activated

        print("Using observation days: {}".format(days_of_runs))
        print("Using duty factors: {}".format(duty_factors))
        print("Using probability of run: {}".format(self.prob_of_run))
        print("Using detectors: {}".format(detectors))
        print("Using Tobs (years): {}".format(self.Tobs_year))
        if 'O4' in self.active_runs:
            print("Using O4 sensitivity: {}".format(self.psd_opts))
        print("Using priors limits: {}".format(self.priors_limits))
        if self.dLmax_depends_on_m1: print("Priors params dLmax(m1): {}".format(self.dLmax_m1_params))
        # define the range for the seed before generation noise in the injections
        self.seedmin = 0
        self.seedmax = 2**32-1
        
    def set_priors(self):

        # create priors once for all, in the global parameter space 'G'
        # this prior is the same for all runs O1, O2, O3, O4
        # random variables are m1, (m2|m1) and dL independent of m1
        # these priors will be used to compute ini_prob of individual injections
        priors_G = self.generate_priors_G()
        self.priors_dict_G = {}
        for k in self.active_runs:
            self.priors_dict_G[k] = priors_G # same priors for all runs in the global space

        if not self.dLmax_depends_on_m1:
            # there is no reduced parameter space, we draw in G
            self.priors_dict_R = self.priors_dict_G

        else:
            # set the priors in the reduced parameter space
            # random variables are m1, (m2|m1) and (dL|m1)
            # if dLmax depends on m1, we use the dLmax(m1,Oi) value, Oi = O1, O2, O3, O4low, O4high
            # empirical functions obtained with simulations using a configuration that maximizes the GW signal
            # m1 = m2, a1 = a2 = 1, inclination = 0...
            # dLmax(m1) = (a[0]+a[1]*m1)*exp(-(m1+a[2])**2/(2*a[3]))
            self.priors_dict_R = {}
            self.dLmax_m1_params = {}
            if self.dLmax_m1_params_args != None: # the user provided the values a[0], a[1], a[2], a[3] for each LVCrun
                for k in self.active_runs: # keep only those for our actual LVCruns
                    self.dLmax_m1_params[k] = self.dLmax_m1_params_args[k]
              
            else: # use the computed ones by simulation
                dLmax_m1 = get_dLmax_params(self.SNR_th) # get params for all runs (O1, O2, O3, O4low, O4high)
                if self.psd_opts == 'low': # rename the O4low or O4high as O4, keep the correct one
                    dLmax_m1['O4'] = dLmax_m1['O4low']
                else:
                    dLmax_m1['O4'] = dLmax_m1['O4high']
                for k in self.active_runs: # keep only active runs
                    self.dLmax_m1_params[k] = dLmax_m1[k]
                        
            for k in self.active_runs:
                self.priors_dict_R[k] = self.generate_priors_R(k)


    def get_condition_func_dl_m1(self):

        return dict(O1=self.condition_func_dlm1_O1,
                    O2=self.condition_func_dlm1_O2,
                    O3=self.condition_func_dlm1_O3,
                    O4=self.condition_func_dlm1_O4)
        
    def condition_func_dlm1_O1(self, reference_params, mass_1):
        
        return dict(minimum=self.priors_limits['dLmin'],maximum=self.getdLmax('O1',mass_1))

    def condition_func_dlm1_O2(self, reference_params, mass_1):

        return dict(minimum=self.priors_limits['dLmin'],maximum=self.getdLmax('O2',mass_1))

    def condition_func_dlm1_O3(self, reference_params, mass_1):

        return dict(minimum=self.priors_limits['dLmin'],maximum=self.getdLmax('O3',mass_1))

    def condition_func_dlm1_O4(self, reference_params, mass_1):

        return dict(minimum=self.priors_limits['dLmin'],maximum=self.getdLmax('O4',mass_1))

    def condition_func_m1m2(self,reference_params, mass_1):

        return dict(minimum=self.priors_limits['Mmin_det'], maximum=mass_1)

    def reduced_space_function(self,x,LVCrun):

        '''
        This function returns pG(m1) * int_lmin^lmax(m1) pG(l) dl which is proportional to pR(m1), 
        i.e. proportional to the marginal distribution of m1 in the reduced parameter space
        x is mass_1 in the detector frame and must be a scalar
        LVCrun is O1 or O2, or O3, or O4
        we use the object self.priors_dict_G that contains the pdf of m1 and luminosity_distance
        in the global parameter space
        '''
        dlmax_m1 = self.getdLmax(LVCrun,x)
        int_l = integrate.quad(self.priors_dict_G[LVCrun]['luminosity_distance'].prob,
                               self.priors_dict_G[LVCrun]['luminosity_distance'].minimum,
                               dlmax_m1)[0]
        return self.priors_dict_G[LVCrun]['mass_1'].prob(x)*int_l

    def getdLmax(self,LVCrun,m1):

        '''
        returns the value dLmax(m1) = (a[0]+a[1]*m1)*exp( -(m1+a[2])^2 / (2*a[3]^2) )
        '''
        dlp = self.dLmax_m1_params[LVCrun]
        dl = (dlp[0]+dlp[1]*m1)*np.exp( - (m1+dlp[2])**2 / (2 * dlp[3]**2) )
        #dlcst = {'O1':8e3,'O2':1e4,'O3':1.3e4}
        #return np.min([dl,dlcst[LVCrun]])
        return dl

    
    def generate_priors_detector_G(self):

        '''
        Define the priors in the detector frame, in the global parameter space
        '''
        
        prior_dict = bl.gw.prior.BBHPriorDict()
        prior_dict.pop('mass_1')
        prior_dict.pop('mass_2')
        prior_dict.pop('mass_ratio')
        prior_dict.pop('chirp_mass')
        prior_dict.pop('luminosity_distance')
        prior_dict['mass_1'] = bl.core.prior.PowerLaw(alpha=self.priors_limits['power_index_m1_det'],
                                                      minimum=self.priors_limits['Mmin_det'],
                                                      maximum=self.priors_limits['Mmax_det'],
                                                      name='mass_1')

        prior_dict['mass_2'] = bl.core.prior.ConditionalPowerLaw(name="mass_2",
                                                                 condition_func=self.condition_func_m1m2,
                                                                 alpha=self.priors_limits['power_index_m2_det'], 
                                                                 minimum=self.priors_limits['Mmin_det'],
                                                                 maximum=self.priors_limits['Mmax_det'],
                                                                 latex_label="mass_2")
        
        # default dL law with a constant dLmax (the same for all runs O1, O2, O3, O4)
        prior_dict['luminosity_distance'] = bl.core.prior.PowerLaw(alpha=self.priors_limits['power_index_dL'],
                                                                   minimum=self.priors_limits['dLmin'],
                                                                   maximum=self.priors_limits['dLmax'],
                                                                   name='luminosity_distance')

        prior_dict['geocent_time'] = bl.core.prior.Uniform(minimum=0.,
                                                           maximum=86400.,
                                                           name='geocent_time')

        return prior_dict
    
    
    def generate_priors_detector_R(self,LVCrun):

        '''
        Define the priors in the detector frame, in the reduced parameter space
        '''

        bdict = bl.gw.prior.BBHPriorDict()
        bdict.pop('mass_1')
        bdict.pop('mass_2')
        bdict.pop('mass_ratio')
        bdict.pop('chirp_mass')
        bdict.pop('luminosity_distance')
        prior_dict = bl.core.prior.ConditionalPriorDict()
        prior_dict.update(bdict.copy())
        prior_dict['geocent_time'] = bl.core.prior.Uniform(minimum=0.,
                                                           maximum=86400.,
                                                           name='geocent_time')
        # then prepare specific priors for m1, m2, dL

        npts = 1000
        vx = np.logspace(np.log10(self.priors_limits['Mmin_det']),np.log10(self.priors_limits['Mmax_det']),npts)
        #vx = np.linspace(self.priors_limits['Mmin_det'],self.priors_limits['Mmax_det'],npts)
        vy = np.zeros(npts)
        for i in np.arange(npts):
            vy[i] = self.reduced_space_function(vx[i],LVCrun) # will be normalized inside bl.core.prior.Interped
            
        # then create pR(m1) with the numerical values (vx,vy), pR(m1) is a pdf as bilby normalizes it
        prior_dict['mass_1'] = bl.core.prior.Interped(name="mass_1",
                                                      xx = vx,
                                                      yy = vy,
                                                      minimum=self.priors_limits['Mmin_det'],
                                                      maximum=self.priors_limits['Mmax_det'],
                                                      latex_label="mass_1")
            
        prior_dict['mass_2'] = bl.core.prior.ConditionalPowerLaw(name="mass_2",
                                                                 condition_func=self.condition_func_m1m2,
                                                                 alpha=self.priors_limits['power_index_m2_det'], 
                                                                 minimum=self.priors_limits['Mmin_det'],
                                                                 maximum=self.priors_limits['Mmax_det'],
                                                                 latex_label="mass_2")
        
        prior_dict['luminosity_distance'] = bl.core.prior.ConditionalPowerLaw(name="luminosity_distance",
                                                                              condition_func=self.dl_m1_functions[LVCrun],
                                                                              alpha=self.priors_limits['power_index_dL'],
                                                                              minimum=self.priors_limits['dLmin'],
                                                                              maximum=self.priors_limits['dLmax'],
                                                                              latex_label="luminosity_distance")
        return prior_dict
        
 
    
    def generate_priors_source(self): # not used in the current version

        '''
        Define the priors in the source frame
        '''
        
        if priors == None:
            prior_dict = bl.gw.prior.BBHPriorDict()
            prior_dict.pop('mass_1')
            prior_dict.pop('mass_2')
            prior_dict.pop('mass_ratio')
            prior_dict.pop('chirp_mass')
            prior_dict.pop('luminosity_distance')

            prior_dict['mass_1'] = bl.core.prior.PowerLaw(alpha=self.priors_limits['power_index_m1_source'],
                                                          minimum=self.priors_limits['Mmin_source'],
                                                          maximum=self.priors_limits['Mmax_source'],
                                                          name='mass_1')
            prior_dict['mass_2'] = bl.core.prior.PowerLaw(alpha=self.priors_limits['power_index_m2_source'],
                                                          minimum=self.priors_limits['Mmin_source'],
                                                          maximum=self.priors_limits['Mmax_source'],
                                                          name='mass_2')
            #prior_dict['redshift'] = bl.gw.prior.UniformComovingVolume(minimum=0.01, maximum=4, name="redshift")
            prior_dict['redshift'] = bl.core.prior.PowerLaw(alpha=self.priors_limits['z_alpha'],
                                                            minimum=self.priors_limits['zmin'],
                                                            maximum=self.priors_limits['zmax'],
                                                            name='redshift')
            prior_dict['luminosity_distance'] = 1

        
        return prior_dict

    
    def do_injections(self,
                      Nsamps=100,
                      output_dir='./injection_files',
                      run=1):


        '''
        This is the function computing the injections
        injections parameters are computed in the detector frame
        it splits the work between child processes
        each of them call the function run_bilby_sim to get the SNR
        the functions defines a shared object (using the class multiprocess/Manager) in the list format (it cannot be a dict)
        at the end of the process the shared object is converted into a dict which will be used afterwards to merge with many
        others to have a large number of injections
        parameters
        ----------
        Nsamps: int
        number of requested accepted injections
        default: 100

        output_dir: str
        name of the output directory where injections file will be stored
        default: 'injection_files'

        run: int
        run id to distinguish injection files when using dag
        default: 1
        '''
        
        if not os.path.isdir(output_dir): os.mkdir(output_dir)
        self.Nsamps = Nsamps # number of detected injections required
        if self.Nsamps > 1000 and self.dump_inj_period<self.Nsamps/50 :
            oldv = self.dump_inj_period
            self.dump_inj_period = int(self.Nsamps/50)
            print("Ask for {} injections: reduce the pickle dump period from: {} injections to {} injections".format(self.Nsamps,oldv,self.dump_inj_period))
        
        self.run = run # keyword, will be added to the output filename
        #self.start_time = int(time.time())
        #pattern = "detected_events_{:d}_{:d}_{:.2f}.p".format(self.start_time,self.run,self.SNR_th)
        pattern = "detected_events_{:d}_{:.2f}.p".format(self.run,self.SNR_th) # pattern of the final dict filename
        self.output_file = output_dir+"/tmp_"+pattern  # temporary file
        self.output_file_dict = output_dir+"/"+pattern # final file, containing the dictionnary
        self.Ndet = Counter(0,self.Nsamps) # min, max, to count the number of detected injections
        self.NsimTot = Counter(0) # min, to count the number of tries, whatever the considered run (O1, O2, O3, O4)
        self.Nsim = {k:Counter(0) for k in self.active_runs} # counter specific to the run id (O1, O2, O3, O4)
        self.ErrorInjections = Counter(0) # to count the number of failures in injections
        self.N_no_ifos = Counter(0) # to count the number of time we have no ifos online
        print("Ndet counter: {} -> {}".format(self.Ndet.get_value(),self.Ndet.max_value()))
        print("Nsim counter: {} -> {}".format(self.NsimTot.get_value(),self.NsimTot.max_value()))
        print("Nsim[Oi] counter:")
        for k in self.active_runs:
            print("\t{}:{}".format(k,self.Nsim[k].get_value()))
        print("Temporary events will be written in {}".format(self.output_file))
        print("Final events will be written in {}".format(self.output_file_dict))
                
        # prepare room for the results
        # results objects is a list, not a dict as the multiprocess class does NOT allow dict to be shared objects
        # the results shared object will be converted into a dict at the end of the execution
        self.results = self.manager.list([self.empty]*self.Nsamps)
        indexes = np.arange(self.Nsamps)
        if self.Nsamps < self.nprocs:
            print("Not enough samples to split among processes. Use only 1 process.")
            self.nprocs = 1
        process_indexes = np.array_split(indexes,self.nprocs) # divide the work between subprocesses

        wproc = mp.Process(target=self.write_injections,args=(self.output_file,self.shared_queue))
        procs = [mp.Process(target=self.run_bilby_sim,args=(process_indexes[i],self.results,self.shared_queue)) for i in range(self.nprocs)]

        # start first write process and write the header for this run
        wproc.start()
        time.sleep(0.1)
        self.header_dict = self.get_header()
        self.header = [self.header_msg,self.header_dict]
        self.shared_queue.put(self.header) # first item of the final list
        time.sleep(0.1)

        for p in procs: p.start()
        for p in procs: p.join() # wait for the requested number of detected events to finish

        print("Child processes terminated... Stopping write_injections process.")
        # write footer in pickle file, we need the final statistics
        self.shared_queue.put([self.footer_msg,self.get_footer()])
        time.sleep(1)
        # then kill the writer
        self.shared_queue.put([self.kill_msg])
        wproc.join()

        if self.Ndet.get_value() > 0:
            print("and dealing with results...")

            # convert the big self.results object into a dict
            self.dict_detected = write_result_to_dict(self.results)

            # add header and footer data
            self.dict_detected.update(self.header_dict.copy())
            self.dict_detected.update(self.get_footer().copy()) # final values of Nsim etc

            # add simulation information, final values
            self.dict_detected['Ndet'] = np.max(self.dict_detected['Ndet'])
            self.dict_detected['NsimTot'] = np.max(self.dict_detected['NsimTot'])
            
            # and write it to file. This is NOT the icarogw/gwcosmo dict at this level
            # the icarogw/gwcosmo dict is created when running the executable create_injection with option --combine: doing this,
            # you merge several dicts, needed to have a large number of injections
            write_object_to_file(self.dict_detected,self.output_file_dict)
            # last check:
            nvals = 1
            if self.Ndet.get_value() > 1: nvals = len(self.dict_detected['SNR'])
            print("NsimTot: {}, Ndet: {}, dict_size: {}, ErrorInjections: {}, #no ifos online: {}"
                  .format(self.NsimTot.get_value(),
                          self.Ndet.get_value(),
                          nvals,
                          self.ErrorInjections.get_value(),
                          self.N_no_ifos.get_value()))
            sim_values = 0
            for k in self.active_runs:
                print("Nsim[{}]: {}".format(k,self.Nsim[k].get_value()))
                sim_values += self.Nsim[k].get_value()
            isok = (sim_values == self.NsimTot.get_value())
            print("Check: NsimTot should be: {}: {} ===> {}".format(sim_values,self.NsimTot.get_value(),isok))
        else:
            print("No injections computed, no output file produced.")
        

        print("Done.")
 

    def get_Nsim(self):
        # return dict of real-time values
        return {k:self.Nsim[k].get_value() for k in self.active_runs}
        
    def get_header(self):

        '''
        This funtion returns a dictionnary containing the common data for all injections
        '''
        hdict  = {'SNR_th':self.SNR_th,
                  'Tobs':self.Tobs_year,
                  'frame':self.frame,
                  'prob_of_run':self.prob_of_run}

        if self.frame == "source":
            hdict['H0_ref'] = self.H0_ref
            hdict['Om0_ref'] = self.Om0_ref
            
        return hdict

    def get_footer(self):

        '''
        this function return the desired information in the footer
        it is set as the last line of temporary files and in the last line of the final dict
        we return the current values of the Nsims for the active runs in case we need to build an injection file before the end of the process
        '''
        tmpdict = dict(Nsim=self.get_Nsim())
        tmpdict['tot'] = np.sum(list(tmpdict['Nsim'].values()))
        return tmpdict # format 'Nsim': {'O1': 223, 'O2': 465, 'O3': 551, 'O4': 517, 'tot':1756}
    
    def write_injections(self,filename,queue):

        '''
        This is the function started by the main process as a child process
        it listens for new data on the socket (queue). New data can be new detected injections
        from the run_bilby_sim function, of a request for auto-kill if the number of requested injections is reached,
        or a header describing the conditions of the injections (SNRth, H0, Om...)
        '''
        writer_start = time.time()
        ngets = 0
        nevts = 0
        nhdr = 0
        nfoot = 0
        nkill = 0
        evt_list = []
        write_counter = 0
        write_done = 0
        while True:
            #print("Wait for something in the queue...")
            item = queue.get() # blocking, listens for new data in the socket
            ngets += 1 # new data arrived            
            if item[0] == self.injection_msg:
                # in these cases we record the data
                if item[0] == self.injection_msg:
                    nevts += 1
                    write_counter += 1
                    evt_list.append(item)
                    if write_counter == self.dump_inj_period:
                        write_counter = 0
                        try:
                            write_done += 1
                            start_0 = time.time()
                            # the values of Nsim Oi are at the time of a detected inj so it's OK
                            # careful: the sum of NOi can differ a bit from the NsimTot of the injection as there are some computed simulations
                            # between the time of the received injection and the time we get the counter in the line just below (get_footer()):
                            footer = [self.footer_msg,self.get_footer()]
                            #evt_list.append(footer) # add temporary the footer to ease further file reading
                            data_to_write = copy.deepcopy(evt_list)
                            data_to_write.append(footer)
                            write_object_to_file(data_to_write,self.output_file)
                            del data_to_write
                            #evt_list.pop() # remove the footer for the next injections
                            start_1 = time.time()
                            print("{} (+{} s) Dump injections in {}, iter #{} ({} injections). Dump time: {} s."
                                  .format(start_0,start_0-writer_start,self.output_file,write_done,nevts,start_1-start_0))
                            sys.stdout.flush()
                        except Exception as e:
                            print(e)
            elif item[0] == self.header_msg:
                evt_list.append(item) # record information in tmp file
                nhdr += 1
            elif item[0] == self.footer_msg: # last message, when normal termination
                evt_list.append(item) # record information in tmp file
                nfoot += 1
            elif item[0] == self.kill_msg:
                nkill += 1
                break # don't listen anymore
            else:
                print("Got unknown message type. Ignoring it.")            

        print("write_injections process terminated, got {} msgs:\n\t{} header\n\t{} injections\n\t{} footer\n\t{} kill".format(ngets,nhdr,nevts,nfoot,nkill))

        # last dump on disk to be sure to have all events
        try:
            print("Last dump of event list on file {}".format(self.output_file))
            write_object_to_file(evt_list,self.output_file)
            write_done += 1
            print("File written ({} times)".format(write_done))
        except Exception as e:
            print(e)

        return
        
    
    def run_bilby_sim(self,n,results,shared_queue):

        '''
        The actual injections are computed in this function
        this is a child process started by the main
        the idea is to draw a detector configuration (O1, O2, O3 and available interferometers)
        the presence of a given interferometer is considered independent on the others
        Parameters
        ----------
        n: is the list of indices attributed to the process to write in the list self.results
        results: is the shared object self.results
        shared_queue: is the communication pipe to the process "write_injections"
        '''
        
        mypid = os.getpid()
        np.random.seed(None) # to have independent samples in the different
        print("child process pid: {}, niter: {}, range: [{}-{}]".format(mypid,len(n),n[0],n[-1]))
        idx_n = 0
        dtloop = 0.
        dtinj = 0.
        ninj = 0
        nloop = 0
        nsim_child_tot = 0 # no reset
        nsim_child_tot_LVC = {k:0 for k in self.active_runs} # no reset
        n_nodets = 0
        while idx_n < len(n): # store results in self.results[idx_n], repeat until requested number is obtained

            start_0 = time.time()

            # increase the number of simulations, event if there are no available interferometers
            nsim_child_tot += 1            
            run_LVC, dets = self.detector_config.GetDetectorConfig() # draw a random detector configuration ([O1, O2, O3, O4], + [L, V, H])
            nsim_child_tot_LVC[run_LVC] += 1
            nsim_global_tot = self.NsimTot.increment() # return value after increment
            nsim_global_tot_LVC = self.Nsim[run_LVC].increment() # return value after increment
            
            if len(dets) == 0: # no need to go further, redraw a new configuration after increase of nsim
                n_nodets +=1
                self.N_no_ifos.increment()
                continue
            
            # load the correct prior dict according to O1, O2, O3, O4
            # self.priors_dict_R is self.priors_dict_G if dLmax independent of m1
            priors_dict = self.priors_dict_R[run_LVC]

            injection_parameters = priors_dict.sample(1)
            injection_parameters = {var: injection_parameters[var][0] for var in injection_parameters.keys()} # change array size 1 values into scalar

            if self.frame == 'source_frame': # compute values in the source_frame
                z = injection_parameters['redshift']
                injection_parameters['luminosity_distance'] = bl.gw.conversion.redshift_to_luminosity_distance(z,cosmology=self.cosmo.name)
                injection_parameters['mass_1'] *= (1+z)
                injection_parameters['mass_2'] *= (1+z)
                    
            duration = np.ceil(gwutils.calculate_time_to_merger(frequency = self.fmin,
                                                                mass_1 = injection_parameters['mass_1'],
                                                                mass_2 = injection_parameters['mass_2']))
            if duration<1: duration=1.

            start_1 = time.time()
                
            waveform_arguments = dict(waveform_approximant = self.approx, reference_frequency = self.fmin, minimum_frequency = self.fmin)
            waveform_generator = bl.gw.waveform_generator.WaveformGenerator(
                sampling_frequency = self.fsamp, duration=duration+1,
                frequency_domain_source_model = bl.gw.source.lal_binary_black_hole,
                parameter_conversion = bl.gw.conversion.convert_to_lal_binary_black_hole_parameters,
                waveform_arguments = waveform_arguments)
            
            ifos = bl.gw.detector.InterferometerList(dets)
            for i in range(len(ifos)):
                ifos[i].power_spectral_density = self.bilby_psds[run_LVC][ifos[i].name]
                
            # draw a random seed and record it if the injection is detected
            #seed_value = np.random.randint(self.seedmin,self.seedmax,size=1,dtype=np.uint32)[0]
            #np.random.seed(seed_value) # to have independent samples in the different processes
            ifos.set_strain_data_from_power_spectral_densities(
                sampling_frequency = self.fsamp, duration=duration+1,
                start_time = injection_parameters['geocent_time']-duration)
            try:
                ifos.inject_signal(waveform_generator = waveform_generator,
                                   parameters = injection_parameters)                
            except Exception as e:
                print(e)
                self.ErrorInjections.increment() # increment shared counter
                continue

            end_1 = time.time()
            dtinj += end_1-start_1
            ninj += 1

            sum_SNR_sq = 0
            for ifo_string in dets:
                sum_SNR_sq += np.real(ifos.meta_data[ifo_string]['matched_filter_SNR'])**2

            SNR = np.sqrt(sum_SNR_sq)

            if SNR >= self.SNR_th: # detected injection

                # we increment here by +nsims, as a new SNR is computed, avoid updating the shared counter at each injection
                # it means that in the temporary injection file we'll always have the correct Ndet and Nsim values
                ndet = self.Ndet.increment() # update Ndet counter
                idx_result = n[idx_n] # write in this place in the shared list

                # compute the draw probability in the global parameter space
                # self.priors_dict_G[run_LVC] are all the same, independently of O1, O2, O3, O4
                #if self.frame == 'source_frame': 
                #    injection_parameters['mass_1'] /= (1+z)
                #    injection_parameters['mass_2'] /= (1+z)
                #    ini_prob = self.priors_dict_G[run_LVC]['mass_1'].prob(injection_parameters['mass_1'])*\
                #        self.priors_dict_G[run_LVC]['mass_2'].prob(injection_parameters['mass_2'])*\
                #        self.priors_dict_G[run_LVC]['redshift'].prob(z)
                #else:

                ini_prob = self.priors_dict_R[run_LVC].prob({'mass_1':injection_parameters['mass_1'],
                                                             'mass_2':injection_parameters['mass_2'],
                                                             'luminosity_distance':injection_parameters['luminosity_distance']})
            
                # add new keys to injections
                injection_parameters['idx'] = idx_result
                injection_parameters['SNR'] = SNR
                injection_parameters['dt'] = duration
                injection_parameters['pi'] = ini_prob
                injection_parameters['run'] = run_LVC
                injection_parameters['Ndet'] = ndet # write the value recorded previously (at the time of detected injection)
                injection_parameters['NsimTot'] = nsim_global_tot # write the value recorded previously (at the time of detected injection)
                injection_parameters['seed'] = 0#seed_value

                # store data in the list
                results[idx_result] = injection_parameters
                
                idx_n += 1 # increment the index for next detected injection
                shared_queue.put([self.injection_msg,results[idx_result]]) # send the data to the write process which will also add the footer
                
            end_0 = time.time()
            dtloop += end_0-start_0
            nloop += 1
            if 0 and (nloop % 100) == 0:
                print("pid: {}, loop: {} ({}), inject: {} ({}), inj/tot: {} num_nodets: {}".format(mypid, dtloop/(1.*nloop),nloop,
                                                                                                   dtinj/(1.*ninj),ninj,
                                                                                                   (dtinj/(1.*ninj))/(dtloop/(1.*nloop)),n_nodets))
                print("\t-> ",mypid,self.Ndet.get_value(),self.Ndet.max_value(),self.NsimTot.get_value())
                #            if self.Ndet.get_value() >= self.Ndet.max_value():
                #                can_quit.set()
                #                break
                
        print("{}: child {}: process terminated... iter {}/{}, nsim_child_tot: {}, num_no_ifos: {}"
              .format(time.time(),os.getpid(),idx_n,len(n),nsim_child_tot,n_nodets))
        for k in nsim_child_tot_LVC.keys():
            print("\tchild Nsim[{}]:{}".format(k,nsim_child_tot_LVC[k]))
        print("\tchild check: \sum nsim[Oi]: {} == nsimtot: {}".format(np.sum(list(nsim_child_tot_LVC.values())),nsim_child_tot))
            
        return

    
    def combine(self,path='./injection_files',output=None,Tobs=None):

        '''
        This function is typically called after distinct injections are computed i.e.
        we have several dictionnaries corresponding to different runs (for instance with dag)
        the path in argument is the directory where all dictionnaries to be merged are located
        output is the filename where the merged dict will be stored
        '''
        
        files = []
        
        for file in os.listdir(path):
            if file.endswith('.p'): # select *.p files (the dict must end with .p)
                files.append(os.path.join(path, file))

        print("Got {} \".p\" file(s) in dir {}".format(len(files),path))
        first_dict = True
        dict_detected = {}
        ndicts = 0
        avg_ratio = 0.
        sq_ratio = 0.
        ndet_loop = 0
        nsim_loop = 0
        for i,path in enumerate(files):
            if os.path.getsize(path) == 0:
                print("Empty pickle file {}, skip it.".format(path))
                continue
            try:
                data = pickle.load(open(path,"rb"))
            except:
                print("Could not load pickle file {}, skip file.".format(path))
                continue
            if not isinstance(data,dict):
                print("Skip not dict file {}".format(path)) # deal with dict file only
                continue
            else:
                ndicts += 1 
                if first_dict: # first open dict, we create its structure
                    first_dict = False
                    for var in data.keys():
                        dict_detected[var] = data[var]
                else: # for the other dicts, we append their data
                    for var in data.keys():
                        dict_detected[var] = np.hstack((dict_detected[var],data[var]))
                ndet_loop += data['Ndet']
                nsim_loop += data['NsimTot']
                print("dict file {} data recorded (#{}), #injections: {}, #sims: {}, acc rate: {} %, Ndet_tot: {}, Nsim_tot: {}.".format(path,ndicts,data['Ndet'],
                    data['NsimTot'],100*data['Ndet']/data['NsimTot'],ndet_loop,nsim_loop))
                ar = 100.*data['Ndet']/data['NsimTot'] # average acceptance in %
                avg_ratio += ar
                sq_ratio += ar**2

        if ndicts == 0:
            print("No valid dict file found, nothing to merge. Exiting.")
            sys.exit()
            
        dict_detected['avg_ratio'] = avg_ratio/ndicts # average acceptance ratio
        if ndicts > 1:
            # stdev of the acceptance ratio
            dict_detected['std_ratio'] = np.sqrt((sq_ratio/ndicts-dict_detected['avg_ratio']**2)/(ndicts-1))
        else:
            dict_detected['std_ratio'] = 0
        print("average Ndet/Nsim in % over {} files: {} {}".format(ndicts,dict_detected['avg_ratio'],dict_detected['std_ratio']))

        # compute the combined number of detected injections and simulated
        dict_detected['Ndet_tot'] = np.sum(dict_detected['Ndet'])
        dict_detected['NsimTot_total'] = np.sum(dict_detected['NsimTot'])

        if ndicts > 1:
            dict_detected['SNR_th'] = dict_detected['SNR_th'][0]
            dict_detected['frame'] = dict_detected['frame'][0]        

        if (Tobs == None) and (ndicts > 1): 
            dict_detected['Tobs'] = dict_detected['Tobs'][0]
        else:
            dict_detected['Tobs'] = dict_detected['Tobs']

        if output == None:
            output = 'combined_pdet_SNR_'+str(dict_detected['SNR_th'])+'.h5' # hdf5 format by default

        # compute the rescaled ini_prob, needed for the injection object used for the analyses
        if ndicts > 1:
            dict_detected['NsimCombined'] = {k:0 for k in dict_detected['prob_of_run'][0].keys()} # combined values of Nsim['Oi']
            prob_of_run = dict_detected['prob_of_run'][0]
        else:
            dict_detected['NsimCombined'] = {k:0 for k in dict_detected['prob_of_run'].keys()} # combined values of Nsim['Oi']
            prob_of_run = dict_detected['prob_of_run']

        # get Nsim for each run Oi for rescaling    
        active_runs = list(dict_detected['NsimCombined'].keys())
        check_ndet = 0
        for k in active_runs: # loop over the runs O1, O2...
            ww = np.where(dict_detected['run'] == k)[0]
            nww = len(ww)
            check_ndet += nww
            if ndicts == 1:
                dict_detected['NsimCombined'][k] = dict_detected['Nsim'][k] # single dict case
            else:
                dict_detected['NsimCombined'][k] = np.sum([dict_detected['Nsim'][i][k] for i in range(len(dict_detected['Nsim']))]) # sum over the files to combine
                
        print("Check: combined total Nsim[{}]: {}".format(k,nww,dict_detected['NsimCombined'][k]))
        print("Got {} dicts for rescaling of probabilities.".format(ndicts))    
        rescale_initial_probabilites(dict_detected,prob_of_run)
        
        # create the injection object used by icarogw or gwcosmo
        
        if dict_detected['frame'] == 'detectors_frame':
            detected = injections_at_detector(m1d=dict_detected['mass_1'],
                                              m2d=dict_detected['mass_2'],
                                              dl=dict_detected['luminosity_distance'],
                                              prior_vals=dict_detected['pi_rescaled'],
                                              snr_det=dict_detected['SNR'],
                                              snr_cut=0,
                                              ifar=dict_detected['SNR']*np.inf,
                                              ifar_cut=0,
                                              ntotal=dict_detected['NsimTot_total'],
                                              Tobs=dict_detected['Tobs'])
        else:
            if ndicts > 1:
                dict_detected['H0_ref'] = dict_detected['H0_ref'][0]
                dict_detected['Om0_ref'] = dict_detected['Om0_ref'][0]

            cosmo_ref = standard_cosmology(Omega_m=dict_detected['Om0_ref'])
            detected = injections_at_source(cosmo_ref=cosmo_ref,
                                            H0_ref=dict_detected['H0_ref'],
                                            m1s=dict_detected['mass_1'],
                                            m2s=dict_detected['mass_2'],
                                            z=dict_detected['redshift'],
                                            prior_vals=dict_detected['pi_rescaled'],
                                            snr_det=dict_detected['SNR'],
                                            snr_cut=0,
                                            ifar=dict_detected['det_snr']*np.inf,
                                            ifar_cut=0,
                                            ntotal=dict_detected['NsimTot_total'],
                                            Tobs=dict_detected['Tobs'])
        
        self.write_injections_hdf5(detected,output,prob_of_run)
        
        fp = os.path.splitext(output)
        dfile = "stacked_"+os.path.basename(fp[0])+".p"
        pickle.dump(dict_detected,open(dfile,"wb"))
        print("Stacked data from {} dict files written in {}.".format(ndicts,dfile))

    def write_injections_hdf5(self,inj,output,prob_of_run):

        fp = os.path.splitext(output)
        if fp[1] != '.h5':
            output = fp[0] + ".h5"
            print("Can't save injections object in a pickle file, we use .h5 format. Will save in {}".format(output))

        h = h5py.File(output,"w")
        h.create_dataset('m1d',data=inj.m1d_original)
        h.create_dataset('m2d',data=inj.m2d_original)
        h.create_dataset('dl',data=inj.dl_original)
        h.create_dataset('pini',data=inj.ini_prior_original)
        h.create_dataset('snr',data=inj.snr_original)
        h.create_dataset('Tobs',data=inj.Tobs)
        h.create_dataset('ntotal',data=inj.ntotal_original)
        h.create_dataset('ifar',data=inj.ifar)
        # write prob_of_run
        for k in prob_of_run.keys():
            kname = 'prob_'+k
            h.create_dataset(kname,data=prob_of_run[k])
        h.close()
        file_size = os.path.getsize(output)
        print("gwcosmo injections object created in file {} ({} bytes).".format(output,file_size))
            
        
class detector_config(object):
    
    def __init__(self,prob_of_run,psd_dict):
        
        self.prob_of_run = prob_of_run
        self.keys = list(prob_of_run.keys()) # list of keys ['O1','O2','O3','O4']
        self.prob = list(prob_of_run.values()) # list of probas
        self.psd_dict = psd_dict # dict of the form {'O1': {'H1': {'frequency':.... {'L1':....
        
    def GetDetectorConfig(self):
        # draw among O1, O2, O3, O4
        LVCrun = np.random.choice(self.keys,1,replace=True,p=self.prob)[0]
        
        dets = []
        # draw online interferometers according to their duty cycle
        # we assume they are independent (they are online/offline independently)
        for det in self.psd_dict[LVCrun].keys():
            p = np.random.rand(1)[0]
            if p <= self.psd_dict[LVCrun][det]['duty_cycle'] and self.psd_dict[LVCrun][det]['duty_cycle'] >= 0:
                dets.append(det)

        return LVCrun, dets


def write_result_to_dict(result):
    
    '''
    This function converts the object in memory self.results (which is a big list of list), into a dict
    each element is a list composed of 2 elements: an integer (idx) and a dict containing injection parameters
    '''
    
    # structure of the object 'result': inj_params
    nevt = len(result)
    # deal with first element to copy the structure and first data
    idict = result[0]
    all_keys = list(idict.keys()) # get all injection keys
    for i in range(1,nevt): # 1 -> nevt-1
        ldict = result[i]
        for var in all_keys:
            idict[var] = np.hstack((idict[var],ldict[var]))            

    return idict
    
def convert_temp_file_to_dict(temp_file, header_msg, footer_msg, inj_msg):

    '''
    This function is called when we want to convert a temporary file into a dict for injections.
    This can be needed in case of crash of the code or if we want to get injections while they are still computed, i.e.
    without waiting for the normal code to end.
    The function returns 3 variables: the dictionnary computed from the list in the temp_file, the number of injections inside it,
    and a boolean telling if the returned dict should be ignored or not (it is the case when number of injections = 0)
    '''
    
    # expected format: temp_file is a pickle file containing a list
    # first item of the list is: ['header',header_dict]
    # all other items of the list are like:
    # ['NewInjection',
    # [12,
    # {'dec': -0.800192697726566,
    #  'ra': 0.44453410132332744,
    #  'theta_jn': 0.6127555351480456,
    #  'psi': 2.3751939066928043,
    #  'phase': 0.4103638539474935,
    #  'a_1': 0.6244476326815177,
    #  'a_2': 0.35941957055779633,
    #  'tilt_1': 0.7172080985916205,
    #  'tilt_2': 0.44697288768091553,
    #  'phi_12': 4.604601783970368,
    #  'phi_jl': 1.0209595930844262,
    #  'mass_1': 1.4451083092070696,
    #  'mass_2': 1.4193210843212622,
    #  'luminosity_distance': 44552.937806656075,
    #  'geocent_time': 47273.6088415261,
    #  'idx': 12,
    #  'SNR': 1.3618151878203781,
    #  'dt': 171.0,
    #  'pi': 1.4555244193873278e-05,
    #  'run': 'O3',
    #  'Ndet': 13,
    #  'NsimTot': 14,
    #  'seed': 36474848]]
    # last item is: ['footer',footer_dict]

    file_size = os.path.getsize(temp_file)
    # check if there are data in the file
    if file_size == 0:
        return 0, 0, 1
    try:
        fd = open(temp_file,"rb")
    except Exception as e:
        print(e)
        print("Could not open file {}. Exiting.".format(temp_file))
        sys.exit()
        
    try:
        evt_list = pickle.load(fd)
        full_dict, ninj, toskip = convert_list_to_dict(evt_list,header_msg,footer_msg,inj_msg)
    except Exception as e:
        print(e)
        print("Could not unpickle the file {}, file could be damaged, try to redownload it?\n Skipping file...".format(temp_file))
        full_dict = 0
        ninj = 0
        toskip = 1

    fd.close()
    return full_dict, ninj, toskip

def convert_list_to_dict(evt_list, header_msg, footer_msg, inj_msg):

    '''
    This function is used when reading data from a temporary file containing injections
    it converts the list of injections loaded in memory into a dict
    the list is actually a list of list, the first item being the header of the object and the other ones the injections
    we first store the injection data into a list of list in order to call the function
    write_result_to_dict to avoid having 2 functions decoding the data
    '''
    
    # evt_list is expected to be the list of data written in the temporary file during injections
    # first item is the header:
    # ['header',header_dict]
    # injection_msg = "NewInjection"
    # header_msg = "header"
    # footer_msg = "footer"

    hdict = {}
    fdict = {}
    idict = {}
    ninj = 0
    results = []
    for i in range(len(evt_list)):
        msg = evt_list[i]
        if msg[0] == header_msg: # first get the header from the list and copy it in the dict
            hdict = msg[1] # copy the header dict
        elif msg[0] == footer_msg:
            fdict = msg[1] # copy the footer dict
        elif msg[0] == inj_msg: # deal with injections data, store all injections
            results.append(msg[1]) # build a list of list (injections)
            ninj += 1
        else:
            print("Unknown message type in evt list: got {}.".format(msg[0]))

    # convert the 'results' injection list into a dict
    idict = write_result_to_dict(results)

    # copy header/footer dicts in idict
    idict.update(hdict)
    idict.update(fdict)

    # update maximum values
    idict['Ndet'] = np.max(idict['Ndet'])
    idict['NsimTot'] = np.max(idict['NsimTot'])
    
    toskip = False
    if ninj == 0:
        print("No injection in tmp file. Ignoring.")
        toskip = True

    return idict, ninj, toskip


def merge_tmpfile(file,header,footer,injection):

    '''
    This function considers a large number of temporary injections files, convert them into dicts.
    This is very useful when we use the temporary files from different clusters without waiting for all of injections to finish.
    Then all these dicts should be used by the 'combine' function to create the icarogw/gwcosmo injection object.
    the arguments 'header' and 'footer' are the strings used to recognize them in the tmpfile
    '''
    
    # file contains the list (full path) of the tmp files to be merged
    # first item is the header:
    # ['header',self.SNR_th,self.Tobs_year,self.frame,self.H0_ref,self.Om0_ref]
    # self.injection_msg = "NewInjection"
    # self.header_msg = "header"
    tmpfiles = []
    with open(file,'r') as f:
        for l in f:
            l = l.rstrip()
            tmpfiles.append(l)
        f.close()
        
    if len(tmpfiles) == 0:
        print("No tmp file to merge. Exiting.")
        sys.exit()

    tmpdir = tempfile.TemporaryDirectory().name
    print("Merged dict will be in dir {}".format(tmpdir))
    os.mkdir(tmpdir)
    nc = 0
    ninj_total = 0
    nskip_total = 0
    strlength = 12
    for i,f in enumerate(tmpfiles):
        print(i,f)
        ldict, ninj, toskip = convert_temp_file_to_dict(f,header,footer,injection)
        ninj_total += ninj
        nskip_total += toskip
        if not toskip:
            ostring = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(strlength))
            #tmpdir+"/dict_"+str(1e9*np.random.rand(1)[0])+"_"+os.path.basename(f)
            outfile = tmpdir+"/dict_"+ostring+"_"+os.path.basename(f) # add random value as file name can be duplicated
            write_object_to_file(ldict,outfile) # write the dict on disk
            nc += 1
    if nc >= 1:
        print("{} injection files have been converted to dict and stored in {}. Total of {} injections.".format(nc,tmpdir,ninj_total))

    print("{} injection files have been skipped.".format(nskip_total))
    
          
def write_object_to_file(myobject,filename):

    f = open(filename,"wb")
    pickle.dump(myobject,f)
    f.close()


def rescale_initial_probabilites(dict_detected,prob_of_run):

    '''
    This function uses the information contained in the combined dictionnary to recompute the initial probabilities
    of the injections
    the probabilities must be rescaled as follow: ini_prob *= (NOi/Ntotal)*(TObs/TobsOi) = (NOi/Ntotal)/prob_of_run[Oi]
    the ratio (TObs/TobsOi) is 1/prob_of_run
    '''
        
    dict_detected['pi_rescaled'] = copy.deepcopy(dict_detected['pi'])
    check_ndet = 0
    for k in prob_of_run.keys(): # loop over the runs O1, O2...
        ww = np.where(dict_detected['run'] == k)[0]
        nww = len(ww)
        check_ndet += nww        
        print("Check: combined total Nsim[{}]: {}".format(k,nww,dict_detected['NsimCombined'][k]))
        if nww > 0:
            print("Rescale probabilities for {}: {} events...".format(k,nww))
            # here we rescale the ini_prob
            # ini_prob *= (NOi/Ntotal)*(TObs/TobsOi) = (NOi/Ntotal)/prob_of_run[Oi]
            dict_detected['pi_rescaled'][ww] *= (dict_detected['NsimCombined'][k]/dict_detected['NsimTot_total'])/prob_of_run[k]
            #print((dict_detected['NsimCombined'][k]/dict_detected['NsimTot_total'])/prob_of_run[k])
            #print("prob_of_run {}: {}".format(k,prob_of_run[k]))
            #print("ratio N{}/Ntot: {}".format(k,dict_detected['NsimCombined'][k]/dict_detected['NsimTot_total']))

    if check_ndet == dict_detected['Ndet_tot']:
        print("All initial probs have been rescaled: {} vs {}.".format(check_ndet,dict_detected['Ndet_tot']))
    else:
        print("Anomaly in the counting of detected injections.")



#!/usr/bin/env python3

import numpy as np
import h5py
import math

def get_offset(LOS_catalog):
    diction = eval(LOS_catalog.attrs['opts'])
    return diction["offset"]

def get_z_array(LOS_catalog):
    return LOS_catalog['z_array'][:]

def get_array(LOS_catalog, arr_name):
    offset = get_offset(LOS_catalog)

    arr = LOS_catalog[str(arr_name)][:]
    arr = np.exp(arr)
    arr -= offset

    return arr

def get_empty_catalog(LOS_catalog):
    return get_array(LOS_catalog, "empty_catalogue")

def get_zprior_full_sky(LOS_catalog):
    return get_array(LOS_catalog, "combined_pixels")

def get_zprior(LOS_catalog, pixel_index):
    return get_array(LOS_catalog, str(pixel_index))

def create_zarray(catalog, band):
    """
    Return zarray for given catalog and band.
    Currently implemented: GLADE+ (B, K, W1, bJ)
    """
    if 'GladePlus' in catalog.name:
        if band in ['B', 'K', 'W1']:
            m = 0.9

            z_lin1 = [1e-6, 4e-5]
            sigmaz1 = 1.5e-6

            z_log1 = [4e-5, 5e-3]
            sigmazoverz1 = 2e-2

            z_lin2 = [5e-3, 1]
            sigmaz2 = 9e-5

            z_log2 = [1, 10]
            sigmazoverz2 = 1e-2/2

            nlin1 = n_linear(z_lin1[0], z_lin1[1], sigmaz1, m)
            nlog1 = n_logarithmic(z_log1[0], z_log1[1], sigmazoverz1, m)
            nlin2 = n_linear(z_lin2[0], z_lin2[1], sigmaz2, m)
            nlog2 = n_logarithmic(z_log2[0], z_log2[1], sigmazoverz2, m)

            zarray = np.linspace(z_lin1[0], z_lin1[1], nlin1, endpoint=False)
            zarray = np.append(zarray, np.logspace(np.log10(z_log1[0]), np.log10(z_log1[1]), nlog1, endpoint=False))
            zarray = np.append(zarray, np.linspace(z_lin2[0], z_lin2[1], nlin2, endpoint=False))
            zarray = np.append(zarray, np.logspace(np.log10(z_log2[0]), np.log10(z_log2[1]), nlog2))

        elif band == 'bJ':
            m = 0.9

            z_lin1 = [1e-6, 5e-4]
            sigmaz1 = 1e-5

            z_log1 = [5e-4, 3e-3]
            sigmazoverz1 = 2e-2

            z_lin2 = [3e-3, 1]
            sigmaz2 = 1e-4

            z_log2 = [1, 10]
            sigmazoverz2 = 0.1

            nlin1 = n_linear(z_lin1[0], z_lin1[1], sigmaz1, m)
            nlog1 = n_logarithmic(z_log1[0], z_log1[1], sigmazoverz1, m)
            nlin2 = n_linear(z_lin2[0], z_lin2[1], sigmaz2, m)
            nlog2 = n_logarithmic(z_log2[0], z_log2[1], sigmazoverz2, m)

            zarray = np.linspace(z_lin1[0], z_lin1[1], nlin1, endpoint=False)
            zarray = np.append(zarray, np.logspace(np.log10(z_log1[0]), np.log10(z_log1[1]), nlog1, endpoint=False))
            zarray = np.append(zarray, np.linspace(z_lin2[0], z_lin2[1], nlin2, endpoint=False))
            zarray = np.append(zarray, np.logspace(np.log10(z_log2[0]), np.log10(z_log2[1]), nlog2))
    else:   #If catalog is not implemented, return zarray for GLADE+ K band.
        m = 0.9

        z_lin1 = [1e-6, 4e-5]
        sigmaz1 = 1.5e-6

        z_log1 = [4e-5, 3e-3]
        sigmazoverz1 = 2e-2

        z_lin2 = [3e-3, 1]
        sigmaz2 = 1e-4

        z_log2 = [1, 10]
        sigmazoverz2 = 1e-2/2

        nlin1 = n_linear(z_lin1[0], z_lin1[1], sigmaz1, m)
        nlog1 = n_logarithmic(z_log1[0], z_log1[1], sigmazoverz1, m)
        nlin2 = n_linear(z_lin2[0], z_lin2[1], sigmaz2, m)
        nlog2 = n_logarithmic(z_log2[0], z_log2[1], sigmazoverz2, m)

        zarray = np.linspace(z_lin1[0], z_lin1[1], nlin1, endpoint=False)
        zarray = np.append(zarray, np.logspace(np.log10(z_log1[0]), np.log10(z_log1[1]), nlog1, endpoint=False))
        zarray = np.append(zarray, np.linspace(z_lin2[0], z_lin2[1], nlin2, endpoint=False))
        zarray = np.append(zarray, np.logspace(np.log10(z_log2[0]), np.log10(z_log2[1]), nlog2))
    return zarray



def n_linear(zmin, zmax, sigmaz, m):
    """
    Returns the number of points between zmin and zmax on the linear scale with the condition that there are  m points (on avarage) between z and z+sigmaz.
    """
    return int((zmax-zmin)/(sigmaz/m))

def n_logarithmic(zmin, zmax, sigmazoverz, m):
    """
    Returns the number of points between zmin and zmax on the logarithmic scale with the condition that there are  m points (on avarage) between z and z+sigma
    """
    return int(m*math.log(zmax/zmin, 1+sigmazoverz))



"""
Rachel Gray 2020
"""

class SchechterParams():
    """
    Returns the source frame Schechter function parameters for a given band.

    Note Mstar is here is M* - 5log10(h)

    ugriz parameters are from https://iopscience.iop.org/article/10.1086/375776/pdf
    (tables 1 and 2)
    B parameters are from https://doi.org/10.1046/j.1365-8711.2002.05831.x
    K parameters are from section 2.1 of
    https://iopscience.iop.org/article/10.3847/0004-637X/832/1/39
    (note paper quotes M*=-23.55 but means M*=-23.55 + 5log10(h))
    """

    def __init__(self, band, schech_alpha=None, schech_Mstar=None, schech_Mmin=None, schech_Mmax=None):
        """
        Parameters
        ----------
        band : observation band (B,K,bJ,u,g,r,i,z)
        """

        self.Mstar = None
        self.alpha = None
        self.Mmin = None
        self.Mmax = None

        self.alpha, self.Mstar, self.Mmin, self.Mmax = self.default_values(band) #initialize to default values

        if schech_alpha is not None: #change to input value if provided
            self.alpha=schech_alpha
        if schech_Mstar is not None:
            self.Mstar=schech_Mstar
        if schech_Mmin is not None:
            self.Mmin=schech_Mmin
        if schech_Mmax is not None:
            self.Mmax=schech_Mmax

    def default_values(self, band): 
        if band == 'B':
            return -1.21, -19.70, -22.96, -12.96
        elif band =='bJ':
            # From https://arxiv.org/pdf/astro-ph/0111011.pdf
            # Mmax from Abstract, of the paper above, valid for fit.
            return -1.21, -19.66, -22.00, -16.5
        elif band == 'K':
            # https://iopscience.iop.org/article/10.1086/322488/pdf
            # Mmax from  Fig 3 of the above reference
            return -1.09, -23.39, -27.0, -19.0
        elif band == 'u':                            #These values are actually u', g', r', i', zi, a.k.a. redshifted to the median redshift (0.1) of SDSS.
            return -0.92, -17.93, -21.93, -15.54 
        elif band == 'g':
            return -0.89, -19.39, -23.38, -16.10 
        elif band == 'r':
            return -1.05, -20.44, -24.26, -16.11 
        elif band == 'i':
            return -1.00, -20.82, -23.84, -17.07 
        elif band == 'z':
            return -1.08, -21.18, -24.08, -17.34 
        elif band == 'W1':
            return -1.12, -24.09, -28, -16.6 # https://iopscience.iop.org/article/10.1088/0004-637X/697/1/506/pdf Tab 3 (All 3.6)
                                             # https://arxiv.org/pdf/1702.07829.pdf (Mmin Mmax Fig2)

        else:
            raise Exception("Expected 'W1', B', 'K', 'u', 'g', 'r', 'i' or 'z' band argument")





"""
Module for late-time cosmology calculations.
Modified gravity models are included.
Currently implements only a flat universe.

Anson Chen

Constants
---------
c : speed of light in km/s
Omega_m : matter fraction
H0 : Hubble parameter


"""

import numpy as np
from scipy import integrate
from scipy.interpolate import splrep, splev, RegularGridInterpolator, interp1d
import lal

c = lal.C_SI/1000.  # 2.99792458e+05 # in km/s


def h(z, Omega_m=0.3065, w0=-1., wa=0.):
    """
    Returns dimensionless redshift-dependent hubble parameter.

    Parameters
    ----------
    z : redshift
    Omega_m : matter fraction
    Dynamical dark energy: w(a) = w0+wa(1-a)

    Returns
    -------
    dimensionless h(z) = sqrt(Omega_m*(1+z)^3 + Omega_Lambda
    *(1+z)^[3(1+w0+wa)]*e^[-3*wa*z/(1+z)])
    """
    Omega_Lambda = (1-Omega_m)
    return np.sqrt(Omega_m*(1+z)**3 + Omega_Lambda* (1+z)**(3*(1+w0+wa)) * np.exp(-3*wa*z/(1+z)))


def dcH0overc(z, Omega_m=0.3065, w0=-1., wa=0.):
    """
    Returns dimensionless combination dc*H0/c
    given redshift and matter fraction.

    Parameters
    ----------
    z : redshift
    Omega_m : matter fraction
    Dynamical dark energy: w(a) = w0+wa(1-a)

    Returns
    -------
    dimensionless combination dc*H0/c = \int_0^z dz'/h(z')
    """
    integrand = lambda zz: 1./h(zz, Omega_m, w0, wa)

    if np.size(z)>1:
        if np.size(np.where(z<=0))>0:
            raise ValueError('Negative redshift input!')
        result = np.array([integrate.quad(integrand, 0, zi)[0] for zi in z])
    else:
        if z<=0:
            raise ValueError('Negative redshift input!')
        result = integrate.quad(integrand, 0, z)[0]  # in km/s

    return result


def dLH0overc(z, Omega_m=0.3065, w0=-1., wa=0.):
    """
    Returns dimensionless combination dL*H0/c
    given redshift and matter fraction.

    Parameters
    ----------
    z : redshift
    Omega_m : matter fraction

    Returns
    -------
    dimensionless combination dL*H0/c = (1+z) * \int_0^z dz'/h(z')
    """
    return (1+z)*dcH0overc(z, Omega_m, w0, wa)



class standard_cosmology(object):

    def __init__(self, H0=70, Omega_m=0.3065, w0=-1., wa=0., zmax=10.0, zmin=1.e-5, zbin=5000):

        self.c = c
        self.H0 = H0
        self.Omega_m = Omega_m
        self.w0 = w0
        self.wa = wa
        self.zmax = zmax
        self.zmin = zmin
        self.zbin = zbin
        self.z_array = np.logspace(np.log10(self.zmin), np.log10(self.zmax), self.zbin)

        # Interpolation of z(dL)
        self.dlH0overc_z_arr = np.array([dLH0overc(z, Omega_m=self.Omega_m, w0=self.w0, wa=self.wa)
                        for z in self.z_array])
        self.dlH0overc_of_z = splrep(self.z_array, self.dlH0overc_z_arr)
        self.z_of_dlH0overc = splrep(self.dlH0overc_z_arr, self.z_array)

        # Interpolation of p(z)
        self.log_pz_arr = np.log10(np.array([self.volume_z(z) for z in self.z_array]))
        self.log_pz_of_z = splrep(np.log10(self.z_array), self.log_pz_arr)

    def update_parameters(self,param_dict):
        """
        Update values of cosmological parameters.
        Key in param_dict: H0
        """
        if 'H0' in param_dict:
            if param_dict['H0'] != self.H0:
                self.H0 = param_dict['H0']

    def dl_zH0(self, z):
        """
        Returns luminosity distance given redshift

        Parameters
        ----------
        z : redshift

        Returns
        -------
        luminosity distance, dl (in Mpc)
        """
        return splev(z, self.dlH0overc_of_z, ext=3)*c/self.H0

    def dgw_z(self, z):
        """
        Returns GW luminosity distance given redshift

        Parameters
        ----------
        z : redshift

        Returns
        -------
        luminosity distance, dgw (in Mpc)
        """
        return self.dl_zH0(z) * self.dgw_dL_ratio(z)

    def z_dlH0(self,dl):
        """
        Returns redshift given luminosity distance

        Parameters
        ----------
        dl : luminosity distance in Mpc

        Returns
        -------
        redshift, z
        """
        return splev(dl*self.H0/c, self.z_of_dlH0overc, ext=3)


    def z_dgw(self,dgw):
        """
        Returns redshift given GW luminosity distance

        Parameters
        ----------
        dgw : GW luminosity distance in Mpc

        Returns
        -------
        redshift, z
        """
        return self.z_dlH0(dgw)

    def dgw_dL_ratio(self, z):
        """
        Returns ratio between GW distance and luminosity distance
        """
        return 1

    def dgw_dL_ratio_dbyz(self, z):
        """
        Returns the derivatives of the ratio between GW distance and
        luminosity distance w.r.t. redshift
        """
        return 0

    def ddgw_dz(self, z):
        """
        Returns the derivative of the GW distance w.r.t. redshift

        Parameters
        ----------
        z : redshift
        """
        H_z = self.H0*h(z, self.Omega_m, self.w0, self.wa)
        dgw_z = self.dgw_z(z)

        return dgw_z/(1+z) + c*(1+z)/H_z * self.dgw_dL_ratio(z) + dgw_z * self.dgw_dL_ratio_dbyz(z)

    def volume_z(self, z):
        """
        Returns the cosmological volume at the given redshift.

        Parameters
        ----------
        z : redshift

        Returns
        -------
        volume element (\int_0^z dz'/h(z'))^2 / h(z): dimensionless
        """
        return dcH0overc(z, self.Omega_m, self.w0, self.wa)**2 / h(z, self.Omega_m, self.w0, self.wa)


    def volume_time_z(self, z):
        """
        Returns the cosmological volume time element at a given redshift.

        Parameters
        ----------
        z : redshift

        Returns
        -------
        volume time element (\int_0^z dz'/h(z'))^2 / (1+z)h(z)
        """
        return self.volume_z(z)/(1.0+z)


    def p_z(self, z):
        """
        p(z|Omega_m): Uniform in comoving volume distribution of galaxies
        """
        return 10.**splev(np.log10(z), self.log_pz_of_z, ext=3)


class local_cosmology(object):

    def __init__(self, H0=70):

        print('Using local cosmology. The result will deviate at high redshift.')

        self.H0 = H0

    def dgw_z(self, z):
        """
        Returns luminosity distance given redshift

        Parameters
        ----------
        z : redshift

        Returns
        -------
        luminosity distance, dl (in Mpc)
        """
        # Local cosmology
        return z*c/self.H0

    def z_dgw(self,dl):
        """
        Returns redshift given luminosity distance

        Parameters
        ----------
        dl : luminosity distance in Mpc

        Returns
        -------
        redshift, z
        """
        # Local cosmology
        return dl*self.H0/c

    def ddgw_dz(self, z):
        """
        Returns the derivative of the luminosity distance w.r.t. redshift

        Parameters
        ----------
        z : redshift
        """
        return c/self.H0

    def p_z(self,z):
        """
        p(z|Omega_m): Uniform in comoving volume distribution of galaxies
        """
        return z*z


class Xi0_n_cosmology(standard_cosmology):
    # Modified cosmology in (Xi0, n) parameterization, inheriting features from standard cosmology.

    def __init__(self, H0=70, Xi0=1, n=1.91, Omega_m=0.3065, w0=-1., wa=0., zmax=10.0, zmin=1.e-5, zbin=5000):

        super().__init__(H0, Omega_m, w0, wa, zmax, zmin, zbin)

        self.H0 = 0
        self.__Xi0 = 0
        self.__n = 0

        self.update_parameters(param_dict={'H0': H0, 'Xi0': Xi0, 'n': n})

    def update_parameters(self, param_dict):
        """
        Update values of cosmological parameters.
        Keys in param_dict: H0, Xi0, n
        """
        if 'H0' in param_dict:
            if param_dict['H0'] != self.H0:
                self.H0 = param_dict['H0']

        update_dgw_dL_ratio = False
        if 'Xi0' in param_dict:
            if param_dict['Xi0'] != self.__Xi0:
                self.__Xi0 = param_dict['Xi0']
                update_dgw_dL_ratio = True
        if 'n' in param_dict:
            if param_dict['n'] != self.__n:
                self.__n = param_dict['n']
                update_dgw_dL_ratio = True
        if update_dgw_dL_ratio:
            self.z_of_dgw = interp1d(self.dgw_dL_ratio(self.z_array)*self.dlH0overc_z_arr, self.z_array, kind='cubic')

    def dgw_dL_ratio(self, z):
        """
        Returns the ratio of GW distance and luminosity distance
        in modified gravity.
        Parametrization introduced by the Geneva group:
        dgw(z) / dL(z) = \Xi(z) = \Xi_0 + (1-\Xi_0)/(1+z)^n

        Parameters
        ----------
        z : redshift
        Xi0, n : modified gravity parameters

        Returns
        -------
        dgw(z) / dL(z)
        """
        return self.__Xi0 + (1-self.__Xi0)/(1+z)**self.__n


    def dgw_dL_ratio_dbyz(self, z):
        """
        Returns the derivative of ratio of GW distance and
        luminosity distance w.r.t. redshift in modified gravity.
        d\Xi(z) / dz = - n * (1-\Xi_0)/(1+z)^(n+1)

        Parameters
        ----------
        z : redshift
        Xi0, n : modified gravity parameters

        Returns
        -------
        d(dgw(z) / dL(z)) / dz
        """
        return -1 * self.__n*(1-self.__Xi0)/((1+z)**(self.__n+1))


    def z_dgw(self, dgw):
        """
        Returns redshift given GW distance

        Parameters
        ----------
        dgw : GW distance in Mpc

        Returns
        -------
        redshift, z
        """
        zgw = self.z_of_dgw(dgw*self.H0/c)

        return zgw



"""
Module for holding all gwcosmo command line arguments
DRY: Don't Repeat Yourself

Rachel Gray
"""

import argparse

class argument:
    def __init__(self, long_flag=None, default=None, metavar=None, type=None, help=None):
        self.long_flag = long_flag
        self.optional_arguments = {
            'default': default,
            'type': type,
            'metavar': metavar,
            'help': help,
        }

def create_parser(*config):
    parser = argparse.ArgumentParser()

    # Each new argument requires only one new line, DRY!
    arguments_dictionary = {
        "--method": argument(None, default=None,
            help="Analysis method, choose from 'gridded' or 'sampling' (required)"),
        "--posterior_samples": argument(None, default=None,
            help="Path to LALinference posterior samples file in format (.dat or hdf5)"),
        "--posterior_samples_field": argument(None, default=None,
            help="Internal field of the posterior samples file, e.g. h5 or json field"),
        "--skymap": argument(None, default=None,
            help="Path to LALinference 3D skymap file in format (.fits or fits.gz)"),
        "--counterpart_ra": argument(None, default=None,
            help="Right ascension of counterpart in radians"),
        "--counterpart_dec": argument(None, default=None,
            help="Declination of counterpart in radians"),
        "--counterpart_z": argument(None, default=None,
            help="Redshift of counterpart (in CMB frame)"),
        "--counterpart_sigmaz": argument(None, default=None,
            help="Uncertainty of counterpart in redshift"),
        "--counterpart_v": argument(None, default=None,
            help="Recessional velocity of counterpart in km/sec (in CMB frame)"),
        "--counterpart_sigmav": argument(None, default=None,
            help="Uncertainty of counterpart in km/sec"),
        "--redshift_evolution": argument(None, default='None',
            help="Allow GW host probability to evolve with redshift. Select between None, PowerLaw or Madau (Default=None)"),
        "--Kcorrections": argument(None, default='False',
            help="Apply K-corrections if true (default=False)."),
        "--reweight_posterior_samples": argument(None, default='True',
            help="Reweight posterior samples with the same priors used to calculate the selection effects (default=True)."),
        "--zmax": argument(None, default=10.0, type=float,
            help="Upper redshift limit for integrals (default=10)"),
        "--galaxy_weighting": argument(None, default='True',
            help="Weight potential host galaxies by luminosity? (Default=True)"),
        "--assume_complete_catalog": argument(None, default='False',
            help="Assume a complete catalog? (Default=False)"),
        "--zcut": argument(None, default=None,
            help="Hard redshift cut to apply to the galaxy catalogue (default=None)"),
        "--mth": argument(None, default=None,
            help="Override the apparent magnitude threshold of the catalogue, if provided (default=None)"),
        "--schech_alpha": argument(None, default=None,
            help="Override the default value for slope of schechter function for given band, if provided (default=None)"),
        "--schech_Mstar": argument(None, default=None,
            help="Override the default value for Mstar of schechter function for given band, if provided (default=None)"),
        "--schech_Mmin": argument(None, default=None,
            help="Override the default value for Mmin of schechter function for given band, if provided (default=None)"),
        "--schech_Mmax": argument(None, default=None,
            help="Override the default value for Mmax of schechter function for given band, if provided (default=None)"),
        "--H0": argument(None, default=70.0, type=float,
            help="Hubble constant value in cosmology (default=70.0)"),
        "--Omega_m": argument(None, default=0.3065, type=float,
            help="Omega matter value in cosmology (default=0.3065)"),
        "--w0": argument(None, default=-1, type=float,
            help="w0 value in cosmology (default=-1.)"),
        "--wa": argument(None, default=0, type=float,
            help="wa value in cosmology (default=0.)"),
        "--nside": argument(None, default=32, type=int,
            help="healpix nside choice for constructing the line-of-sight redshift prior (default=32)"),
        "--coarse_nside": argument(None, default=32, type=int,
            help="coarse healpix nside choice for building the mth and galaxy_norm maps (default=32)"),
        "--maps_path": argument(None, default=None, type=str,
            help="path to store or look for mth and mth maps (default=None)"),
        "--sky_area": argument(None, default=0.999, type=float,
            help="contour boundary for galaxy catalogue method (default=0.999)"),
        "--pixel_index": argument(None, default=None, type=int,
            help="index of the healpix pixel to analyse"),
        "--min_pixels": argument(None, default=30, type=int,
            help="minimum number of pixels desired to cover sky area of event (for use with pixel method only), (default=30)"),
        "--outputfile": argument(None, default='Posterior',
            help="Name of output file (default='Posterior')"),
        "--seed": argument(None, default=None, type=int, 
            help="Random seed (default=None)"),
        "--catalog": argument(None, default=None, metavar="NAME",
            help="""Specify a galaxy catalog by name. Known catalogs are: DESI, DES, GLADE, GLADE+"""),
        "--catalog_band": argument(None, default='B', type=str,
            help = "Observation band of galaxy catalog (B,K,W1,bJ,u,g,r,i,z) (must be compatible with the catalogue provided), (default='B')"),
        "--min_gals_for_threshold": argument(None, default=10, type=int,
            help="The minimum number of galaxies in a coarse pixel for it not to be considered empty (default=10)."),
        "--LOS_catalog": argument(None, default=None,
            help="""The pre-computed line-of-sight redshift prior in hdf5 format"""),
        "--cpus": argument(None, default=1, type=int,
            help="Number of cpus asked for each run (default=1)"),
        "--ram": argument(None, default=1000, type=int,
            help="RAM asked for each run (default=1000 MB)"),
        "--run_on_ligo_cluster": argument(None, default='False', type=str,
            help="Set to true if running on a LIGO cluster (default=False)"),
        "--parameter_dict": argument(None, default=None,
            help="dictionary of parameter set up"),
        "--plot": argument(None, default='True',
            help="Create plots (default=True)"),
        "--sampler": argument(None,
            help="Choice of sampler (choose from Dynesty, Emcee)"),
        "--nwalkers": argument(None, default=100,
            help="Number of walkers for MCMC (default=100)"),
        "--walks": argument(None, default=10,
            help="Minimum number of steps before proposing new live point for nested sampling (10-20 dimensions) (default=10)"),
        "--npool": argument(None, default=2,
            help="Number of pools to use with nested sampling (default=2)"),
        "--nsteps": argument(None, default=1000,
            help="Number of steps for MCMC (default=1000)"),
        "--nlive": argument(None, default=1000,
            help="Number of live points for nested sampling"),
        "--dlogz": argument(None, default=0.1,
            help="Stopping criterion for nested sampling"),
        "--injections_path": argument(None, default=None,type=str,
            help="Path to the injetions file."),
        "--mass_model": argument(None, default=None,type=str,
            help="Mass model. Choose from 'BBH-powerlaw', 'BBH-broken-powerlaw', 'BBH-powerlaw-gaussian', 'NSBH-powerlaw', 'NSBH-broken-powerlaw', 'NSBH-powerlaw-gaussian' or 'BNS'."),
        "--run": argument(None, default=1, type=int,
            help="The number of the run."),
        "--detectors": argument(None, default='HLV', type=str,
            help="The detectors to be cosnidered for the injections (default='HLV')."),
        "--asd_path": argument(None, default='None', type=str,
            help="The path to the asd files' folder. They should be named as detector_run_strain.txt like H1_O1_strain.txt."),
        "--psd": argument(None, default='None', type=str,
            help="Fix to a certain psd: O1, O2, O3 or O4. Default is None which will marginalize over detector sensitivities."),
        "--priors_file": argument(None, default='None', type=str,
            help="Set the path in case of a custom prior file (bilby format). If not the code will use the default priors."),
        "--Nsamps": argument(None, default=100, type=int,
            help="Number of detected events."),
        "--days_of_O1": argument(None, default=129, type=float,
            help="Total days of O1 run (default=129)."),
        "--days_of_O2": argument(None, default=268, type=float,
            help="Total days of O2 run (default=268)."),
        "--days_of_O3": argument(None, default=330, type=float,
            help="Total days of O3 run (default=330)."),
        "--days_of_O4": argument(None, default=330, type=float,
            help="Total days of O4 run (default=330)."),
        "--O4sensitivity": argument(None, default='low', type=str,
            help="Optimistic (high) or pessimistic (low) sensitivity for O4 (default='low')."),
        "--duty_factor_O4_H1": argument(None, default=0.75, type=float,
            help="The duty factor of H1 for the O4 run (default=0.75)."),
        "--duty_factor_O4_L1": argument(None, default=0.75, type=float,
            help="The duty factor of L1 for the O4 run (default=0.75)."),
        "--duty_factor_O4_V1": argument(None, default=0.75, type=float,
            help="The duty factor of V1 for the O4 run (default=0.75)."),
        "--duty_factor_O3_H1": argument(None, default=0.7459, type=float,
            help="The duty factor of H1 for the O3 run (default=0.7459)."),
        "--duty_factor_O3_L1": argument(None, default=0.7698, type=float,
            help="The duty factor of L1 for the O3 run (default=0.7698)."),
        "--duty_factor_O3_V1": argument(None, default=0.7597, type=float,
            help="The duty factor of V1 for the O3 run (default=0.7597)."),
        "--duty_factor_O2_H1": argument(None, default=0.653, type=float,
            help="The duty factor of H1 for the O2 run (default=0.653)."),
        "--duty_factor_O2_L1": argument(None, default=0.618, type=float,
            help="The duty factor of L1 for the O2 run (default=0.618)."),
        "--duty_factor_O2_V1": argument(None, default=0.0777, type=float,
            help="The duty factor of V1 for the O2 run (default=0.0777)."),
        "--duty_factor_O1_H1": argument(None, default=0.646, type=float,
            help="The duty factor of H1 for the O1 run (default=0.646)."),
        "--duty_factor_O1_L1": argument(None, default=0.574, type=float,
            help="The duty factor of L1 for the O1 run (default=0.574)."),
        "--frame": argument(None, default='detectors_frame', type=str,
            help="The frame of the injections(source or detectors) (default is 'detectors_frame')."),
        "--num_threads": argument(None, default=1, type=int,
            help="Number of threads (default is 1)."),            
        "--snr": argument(None, default=9, type=float,
            help="The SNR threshold to claim detection of events (default is 9)."),
        "--fmin": argument(None, default=20, type=float,
            help="The minimum frequency of the waveforms in Hz (default is 20)."),
        "--sampling_frequency": argument(None, default=4096, type=int,
            help="The sampling frequency of the waveforms in Hz (default is 4096)."),
        "--approximant": argument(None, default='IMRPhenomPv2', type=str,
            help="The waveform approximant to be used for the waveform calculation (default is 'IMRPhenomPv2')."),
        "--output_dir": argument(None, default='./injection_files', type=str,
            help="The name of the output folder (default is './injection_files')."),
        "--offset": argument(None, default=1., type=float,
            help="Offset added to redshift priors for saving (default is 1.)"),
        "--tmp_to_dict": argument(None, default=None, type=str,
            help="for injections: converts a tmp file (list format) into a dict format"),
        "--tmp_to_stdout": argument(None, default=None, type=str,
            help="for injections: dump a tmp file (list format) to stdout for control"),
        "--Tobs": argument(None, default=None, 
            help="The total observational time in years."),    
        "--combine": argument(None, default='False', type=str,
            help="Set to True if combining different injections files to one."),
        "--output_combine": argument(None, default=None,
            help="The name of the output file."),
        "--path_combine": argument(None, default='./injection_files', type=str,
            help="The name of the files' folder in case of combining (default is './injection_files'."),
        "--merge_tmpfile_list": argument(None, default=None, type=str,
            help="Combine all files listed in the filename into a unique dict file for injections."),
        "--dLmax_depends_on_m1": argument(None, default=1, type=int,
            help="Uses a max value of luminosity distance depending on m1 (defaut=1) ONLY for SNR in {9, 10, 11, 12}!!!!"),
        
        "--disk": argument(None, default=5000, type=int,
            help="Disk asked for each run (default=5000 MB)"),
        "--search_tag": argument(None, default='ligo.dev.o4.cbc.hubble.gwcosmo', type=str,
            help="Search tag for the runs -- used in LIGO clusters (default=ligo.dev.o4.cbc.hubble.gwcosmo)"),
        "--nruns":argument(None, default=1, type=int,
            help="The total number of runs (default is 1)."),        
            
        "--gravity_model":argument(None, default='GR', type=str,
            help="The gravity model to analyse. Choose between 'GR' and 'Xi0_n' (default='GR')."),

        "--counterpart_dictionary": argument(None, default=None,
            help="Counterpart information for multiple bright sirens"),
        "--post_los": argument(None, default='True', type=None,
            help="posterior samples are conditioned on line-of-sight (default is True); can handle bool and dictionary"),
        "--nsamps": argument(None, default=1000, type=int,
            help="number of samples from posterior, not conditioned over line-of-sight (default is 1000)"),
        "--skymap_prior_distance": argument(None, default="dlSquare", type=str,
            help="Distance prior used when generating the GW skymap. Choose from 'dlSquare', 'Uniform' or 'UniformComoving' (default dlSquare)"), 
        "--skymap_H0": argument(None, default=70, type=float,
            help="H0 when distance prior of skymap is uniform in comoving volume (default is 70)"), 
        "--skymap_Omega_m": argument(None, default=0.3065, type=float,
            help="Omega_m when distance prior of skymap is uniform in comoving volume (default is 0.3065)")
    }

    for arg in config:
        config_instance = arguments_dictionary[arg]

        positional_arguments = [arg]
        if config_instance.long_flag:
            positional_arguments.append(config_instance.long_flag)

        parser.add_argument(*positional_arguments, **config_instance.optional_arguments)

    return parser

#### Full list of command line arguments (for ease of defining config in command line scripts)
"""
"--method", "--posterior_samples", "--posterior_samples_field", "--skymap", "--counterpart_ra", "--counterpart_dec", "--counterpart_z", "--counterpart_sigmaz", "--counterpart_v", "--counterpart_sigmav", "--redshift_evolution", "--Kcorrections", "--reweight_posterior_samples", "--zmax", "--galaxy_weighting", "--assume_complete_catalog", "--zcut", "--mth", "--schech_alpha", "--schech_Mstar", "--schech_Mmin", "--schech_Mmax", "--H0", "--Omega_m", "--w0", "--wa", "--nside", "--coarse_nside", "--maps_path", "--sky_area", "--pixel_index", "--min_pixels", "--outputfile", "--seed", "--catalog", "--catalog_band", "--min_gals_for_threshold", "--LOS_catalog", "--cpus", "--ram", "--run_on_ligo_cluster", "--parameter_dict", "--plot", "--sampler", "--nwalkers", "--walks", "--npool", "--nsteps", "--nlive", "--dlogz", "--injections_path", "--mass_model", "--run", "--detectors", "--asd_path", "--psd", "--priors_file", "--Nsamps", "--days_of_O1", "--days_of_O2", "--days_of_O3", "--days_of_O4", "--O4sensitivity", "--duty_factor_O4_H1", "--duty_factor_O4_L1", "--duty_factor_O4_V1", "--duty_factor_O3_H1", "--duty_factor_O3_L1", "--duty_factor_O3_V1", "--duty_factor_O2_H1", "--duty_factor_O2_L1", "--duty_factor_O2_V1", "--duty_factor_O1_H1", "--duty_factor_O1_L1", "--frame", "--num_threads", "--snr", "--fmin", "--sampling_frequency", "--approximant", "--output_dir", "--offset", "--tmp_to_dict", "--tmp_to_stdout", "--Tobs", "--combine", "--output_combine", "--path_combine", "--merge_tmpfile_list", "--dLmax_depends_on_m1", "--disk", "--search_tag", "--nruns", "--gravity_model", "--counterpart_dictionary", "--post_los", "--nsamps", "--skymap_prior_distance", "--skymap_H0", "--skymap_Omega_m"
"""



"""
Classes to dictate the relationship between GW mergers and host galaxy
properties such as redshift and intrinsic luminosity

Rachel Gray
"""

from .luminosity_function import L_M


class UniformWeighting(object):
    """
    Host galaxy probability relation to luminosity: uniform
    """

    def __init__(self):
        self.luminosity_weights = False

    def model(self, M):
        """
        Uniform weighting

        Parameters
        ----------
        M : float
            absolute magnitude

        Returns
        -------
        float
            1.
        """
        return 1.

    def __call__(self, M):
        return self.model(M)


class LuminosityWeighting(UniformWeighting):
    """
    Host galaxy probability relation to luminosity: proportional to luminosity
    """

    def __init__(self):
        self.luminosity_weights = True

    def model(self, M):
        """
        Luminosity weighting

        Parameters
        ----------
        M : float
            absolute magnitude

        Returns
        -------
        float
            Luminosity
        """

        return L_M(M)



class RedshiftEvolutionConstant(object):
    """
    Parent class for CBC merger rate evolution with redshift, p(s|z)
    
    Each redshift evolution child class must have function called 'model'
    which takes only redshift as an input parameter. All model hyperparameters
    should be defined as attributes to the class
    """
        
    def model(self,z):
        """
        No rate evolution

        Parameters
        ----------
        z : float
            redshift

        Returns
        -------
        float
        """
        return 1.
        
    def __call__(self,z):
        return self.model(z)/(1.+z)


class RedshiftEvolutionMadau(RedshiftEvolutionConstant):
    """
    Merger rate relation to redshift: Madau model
    Equation 2 in https://arxiv.org/pdf/2003.12152.pdf
    Default values are taken from https://arxiv.org/abs/2111.03604
    """
    
    def __init__(self, gamma=4.59, k=2.86, zp=2.47):
    
        """
        Parameters
        ----------
        gamma : float
            Powerlaw index for initial (low z) merger rate, R(z)∝(1+z)^gamma (default=4.59)
        k : float
            Powerlaw index for later (high z) merger rate, R(z)∝(1+z)^-k (default=2.86)
        zp : float
            Peak of merger rate (default=2.47)
        """

        self.gamma = gamma
        self.k = k
        self.zp = zp

    def model(self, z):
        """
        Madau rate evolution

        Parameters
        ----------
        z : float
            redshift

        Returns
        -------
        float
        """

        C = 1+(1+self.zp)**(-self.gamma-self.k)
        return C*((1+z)**self.gamma)/(1+((1+z)/(1+self.zp))**(self.gamma+self.k))


class RedshiftEvolutionPowerLaw(RedshiftEvolutionConstant):
    """
    Merger rate relation to redshift: power-law model
    Default value taken from https://arxiv.org/abs/2111.03604
    """

    def __init__(self, gamma=4.59):
    
        """
        Parameters
        ----------
        gamma : float
            Powerlaw index for merger rate, R(z)∝(1+z)^gamma (default=4.59)
        """

        self.gamma = gamma

    def model(self, z):
        """
        Power-law rate evolution

        Parameters
        ----------
        z : float
            redshift

        Returns
        -------
        float
        """
        return (1+z)**self.gamma





from ligo.gracedb.rest import GraceDb
import argparse

def download_skymap(id, db, args):
	outfile = open("{0}_skymap.fits.gz".format(id), "wb")
	try:
		# Use LALInference skymap if possible
		r = db.files(id, "LALInference.fits.gz")
		if args.verbose:
			print("Using LALInference skymap for {0}".format(id))
	except:
		r = db.files(id, "bayestar.fits.gz")
		if args.verbose:
			print("Using Bayestar skymap for {0}".format(id))

	outfile.write(r.read())
	outfile.close()

def main():
	parser = argparse.ArgumentParser(description = "Download skymaps from a list of events")
	parser.add_argument("event", nargs="+", help = "A list of gravitational-wave events, can be either GID for GW event or SID for superevent")
	parser.add_argument("--verbose", action = "store_true", help = "Be very verbose")
	args = parser.parse_args()
	# FIXME Make sure that you have a valid proxy
	client = GraceDb()

	for event_id in args.event:
		download_skymap(event_id, client, args)
	
if __name__ == "__main__":
	main()



import gwcosmo
import numpy as np
import astropy
import astropy.constants as constants
from astropy.table import Table
from astropy import units as u
import pickle

import h5py
import healpy as hp


GLADE = Table.read("GLADE+_reduced.txt",format='ascii')
print("GLADE has a total of "+str(len(GLADE))+" objects.")

# remove galaxies with no reported redshift
no_zs = GLADE['col9'] == 'null'
GLADE.remove_rows(no_zs)

del no_zs

print("Removed null redshift: GLADE has "+str(len(GLADE))+" galaxies with a measured redshift.")

# remove QSOs and galaxy clusters.
no_clusters = GLADE['col3'] == 'C'
GLADE.remove_rows(no_clusters)
del no_clusters

no_QSOs = GLADE['col3'] == 'Q'
GLADE.remove_rows(no_QSOs)
del no_QSOs

print("Removed Quasars and clusters: GLADE has "+str(len(GLADE))+" galaxies.")

# Make sure redshifts are positive
mask = GLADE['col9'].astype('float64') < 0
GLADE.remove_rows(mask)
print("Removed negative redshift: GLADE has "+str(len(GLADE))+" galaxies.")
del mask

#  Remove galaxies without peculiar velocity corrections unless the galaxy is above redshift 0.05
mask = (GLADE['col10'] == 0) & (GLADE['col9'].astype('float64') <= 0.05)
GLADE.remove_rows(mask)
print("Removed Galaxies without peculiar correction: GLADE has "+str(len(GLADE))+" galaxies.")
del mask

#  Remove galaxies without measured redshift and distance (0) and with redshift from distance measure (2)
mask = (GLADE['col12'] == 0) | (GLADE['col12'] == 2)
GLADE.remove_rows(mask)
print("Removed galaxies without redshift and with redshift from dl: GLADE has "+str(len(GLADE))+" galaxies.")
del mask

GLADE.add_column(np.zeros(len(GLADE)),name='zerror')

z_error_pec = GLADE['col11'].data
mask = z_error_pec == 'null'
z_error_pec[mask] = '0'
z_error_pec=z_error_pec.astype(np.float)
del mask

# Z error for galaxies with spectroscopic errors, this is an absolute error
zerror_sp = 1.5e-4
zs_sp = GLADE['col12'] == 3
print('ciao2',zerror_sp)
GLADE['zerror'][zs_sp] = np.sqrt(zerror_sp**2.+z_error_pec[zs_sp]**2.)

del zs_sp

# Z error for galaxies with WISE, this is a relative error of 4%(From Maciej discussion)
zerror_wise = 0.04
zs_wise = GLADE['col2'] != 'null'
z_wise = GLADE['col9'][zs_wise]
z_wise = z_wise.astype('float64')
GLADE['zerror'][zs_wise] = np.sqrt((zerror_wise*(1+z_wise))**2.+z_error_pec[zs_wise]**2.)

del zs_wise
del z_wise


# Z error for galaxies with photoz this is an absolute error
zerror_ph = 1.5e-2
zs_ph = GLADE['col12'] == 1
GLADE['zerror'][zs_ph] = np.sqrt(zerror_ph**2.+z_error_pec[zs_ph]**2.)

del zs_ph

# Z error for galaxies from Hyperleda, this is a relative error. Since these are small redshift the relative error is multiplied only to z
zerror_hyperleda = 0.36
zs_hyperleda = (GLADE['col1'] != 'null')
z_hyperleda = GLADE['col9'][zs_hyperleda]
z_hyperleda = z_hyperleda.astype('float64')
GLADE['zerror'][zs_hyperleda] = np.sqrt((zerror_hyperleda*z_hyperleda)**2.+z_error_pec[zs_hyperleda]**2.)

del zs_hyperleda
del z_hyperleda

# Select columns of interest
ra = GLADE['col4'].astype('float64')*np.pi/180.
dec = GLADE['col5'].astype('float64')*np.pi/180.
z = GLADE['col9'].astype('float64')
B = GLADE['col6']
K = GLADE['col7']
W1 = GLADE['col8']
sigmaz = GLADE['zerror']

del GLADE

ra = ra.astype('float64')
dec = dec.astype('float64')
z = z.astype('float64')
B = B.data
K = K.data
W1= W1.data

mask = B == 'null'
B[mask] = '0'
del mask
mask = K == 'null'
K[mask] = '0'
del mask
mask = W1 == 'null'
W1[mask] = '0'
del mask

B = B.astype('float64')
K = K.astype('float64')
W1 = W1.astype('float64')

B[B==0] = np.nan
K[K==0] = np.nan
W1[W1==0] = np.nan

import h5py
import healpy as hp

# Save GLADE catalog
bands=['B', 'K','W1']
nGal = len(z)
m = np.column_stack((B,K,W1))

del B
del K
del W1

nside = 1024
ind = hp.pixelfunc.ang2pix(nside,dec+np.pi/2,ra)

ra_dec_lim = 0
ra_min = 0.0
ra_max = np.pi*2.0
dec_min = -np.pi/2.0
dec_max = np.pi/2.0 

with h5py.File("glade+.hdf5", "w") as f:
    f.create_dataset("ra", data=ra)
    f.create_dataset("dec", data=dec)
    f.create_dataset("z", data=z)
    f.create_dataset("sigmaz", data=sigmaz)
    f.create_dataset("skymap_indices", data=ind)
    f.create_dataset("radec_lim", data=np.array([ra_dec_lim,ra_min,ra_max,dec_min,dec_max]))
    for j in range(len(bands)):
        f.create_dataset("m_{0}".format(bands[j]), data=m[:,j])